package PureMessage::Install;
use strict;
use HTTP::Request;
use ActiveState::Unix::Pw qw(useradd userdel groupadd groupdel su);
use ActiveState::Unix::Network qw(interfaces);
use ActiveState::Unix::Crontab qw(cron_allow_user cron_remove_user);
use ActiveState::Handy qw(run);
use Net::Domain qw(hostdomain hostfqdn);
use Socket;                         # for AF_INET
use IO::Socket;
use IO::Socket::INET;
use PureMessage::Install::OptionValidator;
use PureMessage::Install::Log;
use PureMessage::Install::Config;
use PureMessage::Install::Shell;
use PureMessage::Install::Utils qw(passwd get_iq_conf
                                   run_as_pmx pmx_can_run
                                   );
use PureMessage::Install::Upgrade;
use PureMessage::Install::Prefix qw(read_prefix_file);
use PureMessage::Snapshot;
use PureMessage::Manager::PasswordGenerator qw(mk_password);
use File::Copy qw(copy);
use File::Path qw(mkpath rmtree);
use PPM::Validate qw(fetch_file get_no_verification);
use PPM::PureMessage::Utils qw(escrow_dir pmx_user create_tmp_dir tmp_dir_location);
use PureMessage::_sys;
use Digest::SHA1 qw(sha1_hex);
use Data::Dumper;
use Exporter;
use Config;
our @ISA = ('Exporter');
our @EXPORT = qw(eula_accepted uninstall_milter uninstall_pmx_user 
                 remove_prefix installed_roles install_package
                 calc_lic_comp);

my $NO_SNAPSHOT_FILE = '/etc/puremessage-nosnapshot';
# System tools
my $zcat = "zcat";
my $gz = ".Z";
if ($^O eq "linux" || $^O eq "freebsd") {
    $zcat = "gunzip -c";
    $gz = ".gz";
}

my $logstr = " >> " . inst_log() . " 2>&1";

# Documentation of Install Object
#
# pmx_sh - shell installer owned by root
# repo   - repository from shell installer
# location - location flag for PPM
# prefix - installation prefix
#
# role_opts - hash of options that are dependent on selected roles
#             sub-classes may set this to have specific values for certain
#             install options
#             - key is the option name, value is a hashref
#               - hashref contains the option/values that should be used
#
# iq_conf - reference to the parsed IQ file
# 
# licensed_pkgs - packages we have licenses for - comes from "pmx_sh components"
# 
# pkgs - Packages to install - subclasses will set this.  It may be added to via
#        the add_pkg() function
#
# lic_pkgs - Packages to install which we must check are licensed
#
# _rio_hash - Required Install Options - hashref of all of the install options 
#             found in the IQ file
#             - key is option name, value is reference to option hashref from IQ file
#
# _rio_array - An ordered list of all of the install options
#
# roles - the list of roles to be installed
#
# pkg_msgs - a list of info and error messages emitted during package install
#    - each element is a hashref containing "type" and "msg" keys.
#    - type will be either "info" or "err"

sub new {
    my $class = shift;
    my %opts = @_;

    die "repo is mandatory" unless defined $opts{repo};
    $opts{pmx_sh} ||= '';

    my $self = { pmx_sh => $opts{pmx_sh},
                 repo => $opts{repo},
                 prefix => $opts{prefix},
                 log_file => inst_log(),
                 role_opts => {},
                 auto_config => $opts{auto_config},
                 roles => [],
                 pkg_msgs => [],
               };
    inst_log("Creating PureMessage::Install object - pmx_sh=$self->{pmx_sh} "
             . "repo=>$self->{repo} prefix=>$self->{prefix}\n");

    bless $self, $class;
    $self->{licensed_pkgs} = $self->calc_lic_comp();

    eval {
        $self->{iq_conf} = get_iq_conf($self->{repo});
    };
    die <<EOT if $@;
Error while retrieving installation options file:

$@
EOT
    return $self;
}

sub calc_lic_comp {
    my ($prefix, $pmx_sh);
    if (@_ == 1) {
        my $self = shift;
        $prefix = $self->{prefix};
        $pmx_sh = $self->{pmx_sh};
    }
    else {
        my %opts = @_;
        $prefix = $opts{prefix};
        $pmx_sh = $opts{pmx_sh};
    }

    my @extra_pkgs = ();
    if ($prefix and -x "$prefix/bin/perl") {
        my %values = (
            'PureMessage::VirusCheck'	=> 2,
            'PureMessage::SpamCheck'	=> 1,
        );
        # get the command to get the licensed features
        my $cmd = run_as_pmx( _norun => 1, prefix => $prefix,
                      command => <<EOT);
$prefix/bin/perl -MPureMessage::_sys -le 'print for keys \%{{PureMessage::_sys::lic_features(PureMessage::_sys::lic_file())}}' 2>/dev/null
EOT
        my @foo = `$cmd`;
        @extra_pkgs = (
            sort {
                ($values{$b} || 0) <=> ($values{$a} || 0) ||
                $a cmp $b
            }
            map { chomp; $_ }
            grep /^PureMessage::/, @foo);
    }
    else {
        # Compute the list of licensed components.  Note that this list should not
        # be trusted, as it comes from the "sub_comps" variable in the installer.sh
        if (defined $pmx_sh) {
            @extra_pkgs = map { chomp; $_ } split / /, `$pmx_sh components`;
        }
    }
    my %extra;
    for (@extra_pkgs) {
        next unless /^PureMessage::(Policy|ServerGroup)/;
        chomp;
        my $pkg = $_;
        $pkg =~ s/::/-/g;
        $extra{$pkg}++;
    }
    inst_log("Licensed components: (". join(", ", @extra_pkgs) . ")");
    return \%extra;
}

sub get_packages {
    my $self = shift;
    my @pkgs;
    @pkgs = @{$self->{pkgs}} if exists $self->{pkgs};
    $self->{lic_pkgs} ||= [];
    foreach my $p (@{$self->{lic_pkgs}}) {
        # make sure this component is licensed
        next unless $self->{licensed_pkgs}{$p};
        push @pkgs, $p;
    }
    inst_log("Packages: @pkgs");
    return \@pkgs;
}

sub _get_rios {
    my $self = shift;
    $self->{_rio_array} = undef;

    # determine which options we need from our package list
    my %rio_names = ();
    foreach my $pkg ( @{$self->get_packages} ) {
        if (exists $self->{iq_conf}{packages}{$pkg}) {
            foreach my $o (@{$self->{iq_conf}{packages}{$pkg}}) {
                $rio_names{$o}++;
            }
        }
    }

    # Certain roles may request that RIOs get prompted or get specific values
    foreach my $o (keys %{$self->{role_opts}}) {
        unless (exists $self->{iq_conf}{option}{$o}) {
            die "Install error: Can't find install option $o - please "
                . "contact PureMessage Support";
        }
        foreach my $k (keys %{$self->{role_opts}{$o}}) {
            inst_log("Setting role option $o.$k to $self->{role_opts}{$o}{$k}");
            $rio_names{$o}++;
            $self->{iq_conf}{option}{$o}{$k} = $self->{role_opts}{$o}{$k};
        }
    }

    # Parse any options we got from the CSM
    # store these rios seperately until we parse them all
    my %csm_rio_names;
    foreach my $o (keys %{$self->{csm_opts}}) {
        unless (exists $self->{iq_conf}{option}{$o}) {
            inst_log("Found invalid install option - ($o)");
        }
        foreach my $k (keys %{$self->{csm_opts}{$o}}) {
            # if the install option wasn't already going to be asked,
            # it shouldn't be prompted for due to this.
            if (!$rio_names{$o}) {
                inst_log("Setting csm option $o.must_prompt to 0");
                $self->{iq_conf}{option}{$o}{must_prompt} = 0;
            }
            $csm_rio_names{$o}++;
            inst_log("Setting csm option $o.$k to $self->{csm_opts}{$o}{$k}");
            $self->{iq_conf}{option}{$o}{$k} = $self->{csm_opts}{$o}{$k};
        }
    }
    $rio_names{$_}++ foreach keys %csm_rio_names;

    my @rios;
    foreach my $k (keys %rio_names) {
        my $o = $self->_setup_rio($k);
        inst_log("Found install option: $k");
        inst_log("install option $k: " . Dumper($o)) if $ENV{PMX_SETUP_DEBUG};
        push @rios, $o;
    }

    my $i = 100;
    my %sort_order = map { $_, $i-- } @{$self->{iq_conf}{option_order}};
    my $sorter = sub {
        my $sa = $sort_order{$a->{name}} || 0;
        my $sb = $sort_order{$b->{name}} || 0;
        return $sb <=> $sa if $sa or $sb;
        return $b->{name} cmp $a->{name};
    };
    @rios = sort $sorter @rios;
    inst_log("Install options: " . join(", ", map {$_->{name}} @rios));
    push @{$self->{_rio_array}}, @rios;
}

# Calculate the defaults and put the RIO in _rio_hash 
# RIO is NOT added to _rio_array, as it should be sorted properly
sub _setup_rio {
    my ($self, $rio_name) = @_;
    unless (exists $self->{iq_conf}{option}{$rio_name}) {
        die "Install error: Can't find install option $rio_name - please "
            . "contact PureMessage Support";
    }
    my $opt = $self->{iq_conf}{option}{$rio_name};
    $self->{_rio_hash}{$rio_name} = $opt;
    $opt->{name} = $rio_name;
    $opt->{default} ||= "";
    $opt->{prompt_user} = $opt->{must_prompt} unless defined $opt->{prompt_user};
    while($opt->{default} =~ /\%\%([^\%]+)\%\%/) {
        my $d = $self->_calc_default($1);
        $opt->{default} =~ s/\%\%\Q$1\E\%\%/$d/;
    }
    $self->{inst_opts}{$rio_name} ||= $opt->{default};
    return $opt;
}

sub reset_promptedness {
    my ($self, @opts) = @_;

    if (@opts) {
        @opts = grep defined, map {$self->{_rio_hash}{$_}} @opts;
    }
    else {
        @opts = @{$self->{_rio_array}};
    }

    foreach my $r (@opts) {
        inst_log("Reseting promptedness of $r->{name}");
        $r->{prompt_user} = $r->{must_prompt};
    }
}

sub add_pkg {
    my ($self, $pkg) = @_;
    return if grep {$_ eq $pkg} @{$self->{pkgs}};
    $self->_get_rios(); 
    push @{$self->{pkgs}}, $pkg;
    if (exists $self->{iq_conf}{packages}{$pkg}) {
        foreach my $o (@{$self->{iq_conf}{packages}{$pkg}}) {
            inst_log("Adding install option $o for package $pkg");
            push @{$self->{_rio_array}}, $self->_setup_rio($o);
        }
        # Re-get the install options so we can sort them.
        $self->_get_rios();
    }
}

sub get_first_rio {
    my $self = shift;
    $self->_get_rios();
    $self->{current_rio} = -1;
    return $self->get_next_rio();
}

# return the next rio that we need to prompt for
# the must_prompt flag may have changed since we originally calculated it
sub get_next_rio {
    my $self = shift;
    do {
        $self->{current_rio}++;
        my $r = $self->{_rio_array}[$self->{current_rio}];
        return undef unless defined $r->{name};
        inst_log("get_next_rio - current=$self->{current_rio} $r->{name}");
        return $r if $r->{prompt_user};
    } while ($self->{current_rio} < @{$self->{_rio_array}});
    return undef;
}

sub get_prev_rio {
    my $self = shift;
    do {
        $self->{current_rio}--;
        my $r = $self->{_rio_array}[$self->{current_rio}];
        return $r if $r->{prompt_user};
    } while ($self->{current_rio} >= 0);
    return undef;
}

sub get_rio_list {
    my $self = shift;
    my @rios;
    foreach my $r (@{$self->{_rio_array}}) {
        if ($r->{must_prompt}) {
            my $value = $self->{inst_opts}{$r->{name}};
            $value =~ s/./\*/g if $r->{password};
            push @rios, { name => $r->{nice_name},
                          value => $value,
                        };
        }
    }
    return @rios;
}

sub set_instopt {
    my ($self, $opt_name, $value, %subopts) = @_;

    $self->_get_rios unless $self->{_rio_hash} and %{$self->{_rio_hash}};
    my $opt = $self->{_rio_hash}{$opt_name} || 
                die "Error - $opt_name is not a valid option!";

    if (exists $subopts{skip_validation}) {
        delete $opt->{must_match};
        delete $subopts{skip_validation};
        inst_log("Skipping validation for install option $opt_name\n");
    }

    $value ||= "";
    inst_log("Setting install option $opt_name to ($value)\n") 
        unless $opt->{password};
    if (exists $opt->{must_match}) {
        # perform validation
        if ($opt->{must_match} =~ /^qr\/(.+)\/$/) {
            unless ($value =~ $1) {
                die ($opt->{no_match_message} || 
                      "Error - $value is not an acceptable response.");
            }
        }
        elsif ($opt->{must_match} eq "__pmx_user") {
            $value = PureMessage::Install::OptionValidator::valid('user', $value);
            my @pw = getpwnam($value);
            if (@pw) {
                if ($self->{prefix} eq $pw[7]) {
                    die "The home directory of '$value' is the same as the "
                        . "PureMessage installation directory "
                        . "'$self->{prefix}'.  PureMessage requires the "
                        . "PureMessage installation directory to be owned by "
                        . "'root' so this is not permitted.";
                }
                unless(exists $subopts{use_existing}) {
                    return { subopt_name => "use_existing",
                             question => "This user already exists.  Do you "
                                 . "want to use it as the PureMessage user?",
                           };
                }
                elsif (! $subopts{use_existing}) {
                    die "Please choose a different user";
                }
                unless (-d $pw[7] or
                        $pw[7] eq "$self->{prefix}/home") {
                    die <<EOT;
The '$value' user's home directory doesn't exist!

Please choose a user with a valid home directory.
EOT
                }
                my $g = getgrgid($pw[3]);
                unless ($g) {
                    die "Can't determine group $pw[3] for '$value', please "
                        . "choose a different user";
                }
                if (exists $subopts{group_ok}) {
                    if ($subopts{group_ok}) {
                        $self->prompt_user("pmx_password", 0);
                    }
                    else {
                        die "Please choose a different username";
                    }
                }
                if (!$subopts{group_ok} and (1 || $g ne $value)) {
                    return { subopt_name => "group_ok",
                             question => <<EOT,
The primary group for '$value' is '$g'.  All installed files will be made 
readable for users in this group.  The PureMessage log files and any 
quarantined mail will be writable as well as readable for users in this group.  

Is this acceptable?
EOT
                           };
                }
            }
            else {
                # incase we disabled this prompt when the user entered an existing user,
                # and then changed that user later
                $self->prompt_user("pmx_password", 1);
            }
        }
        elsif ($opt->{must_match} =~ /^__(\w+)$/) {
            eval {
                $value = PureMessage::Install::OptionValidator::valid($1, $value);
            };
            if ($@) {
                die ($opt->{no_match_message} || "Error: $@");
            }
        }
        elsif (lc($opt->{must_match}) eq "bool") {
            # do nothing - it is either true or false
        }
        elsif (ref($opt->{must_match}) eq "ARRAY") {
            die "Error - you must choose an option" unless $value;
            my $match;
            foreach my $m (@{$opt->{must_match}}) {
                $match++ if $value eq $m;
            }
            die "Error - $value is not a valid value" unless $match;
        }
        else {
            die "Error - $opt->{must_match} is an unknown validator.";
        }
    }

    # Special cases for specific options
    if ($opt_name eq "mta") {
        if ($value =~ /JSMS/) {
            $value = "JSMS";
            # sendmail and postfix provide a local mail sender on
            # localhost:10026, but JSMS does not, so we must find one
            inst_log("JSMS installed - prompting for mail_sender");
            my $ms = $self->_calc_default("smtp_server");
            if ($ms) {
                $self->set_instopt("mail_sender", $ms);
            }
            else {
                # couldn't find one, so prompt user
                # role_opts is set so it will stick, and we set it in 
                # _rio_hash so it will take effect immediately
                $self->{role_opts}{mail_sender}{must_prompt} = 1;
                $self->{_rio_hash}{mail_sender}{must_prompt} = 1;
                $self->reset_promptedness("mail_sender");
            }
        }
        # Non-JSMS MTAs will require their manager package
        $self->add_pkg("PureMessage-$value");
        inst_log("Added PureMessage-$value to package list");
    }

    $self->{inst_opts}{$opt_name} = $value;
    $value =~ s/./*/g if $opt->{password};
    inst_log("Install option $opt_name set to \"$value\"");
}

sub prompt_user {
    my ($self, $opt, $prompt) = @_;

    die "$opt isn't a valid option" unless exists $self->{_rio_hash}{$opt};
    my $r = $self->{_rio_hash}{$opt};
    $r->{prompt_user} = $prompt if defined $prompt;
    return $r->{prompt_user};
}

sub get_instopt {
    my ($self, $name) = @_;
    my $val = $self->{inst_opts}{$name};
    my $tempval = $val || '';
    $tempval =~ s/./*/g if $name =~ /password/;
    inst_log(qq(Returning inst_opt $name => "$tempval") );
    return $val;
}

sub is_fresh_install {
    my $self = shift;
    # XXX: this is a hack, please strengthen this later:
    # no roles will be specified during an upgrade, but the creator of an
    # instance of this class could actually specify if we're doing an upgrade
    # or not (and, maybe, even "how far" we're upgrading).
    return ($self->{roles} && @{$self->{roles}}) ? 1 : 0;
}

sub _calc_default {
    my ($self, $d) = @_;
    $d = lc($d);
    if ($d eq "prefix") {
        return $self->{prefix};
    }
    if ($d eq "domain") {
        return $self->{inst_opts}{domain} || hostdomain();
    }
    if ($d eq "smtp_server") {
        my $mail_sender = "";
        my $domain = $self->{inst_opts}{domain} || hostdomain();
        foreach my $h ("mail.$domain", "smtp.$domain", "localhost") {
            # check if we can connect to the smtp port
            if (IO::Socket::INET->new(PeerAddr => "$h:25", Timeout => 5)) {
                $mail_sender = "$h:25";
                last;
            }
        }
        return $mail_sender;
    }
    if ($d eq "solaris64") {
        if ($^O eq 'solaris' and $ENV{PMX_SOLARIS64}) {
            my $isainfo = `isainfo 2>/dev/null`;
            if ($? or $isainfo !~ /sparcv9/) {
                $self->prompt_user('solaris64', 1);
                return 0;
            }
        }
        return 0;
    }
    if ($d eq "fq_hostname") {
        my $hostname = lc(hostfqdn());
        return $hostname;
    }
    if ($d eq "postfix_relay_networks") {
        my $default = "127.0.0.0/8";
        my $hostname = lc(hostfqdn()) or return $default;
        my $host = gethostbyname($hostname) or return $default;
        my $addy = inet_ntoa($host) or return $default;
        $addy =~ s/\.\d+$/.0\/24/;
        return "$default, $addy";
    }
    if ($d eq "subnet") {
        my @ifs = interfaces;
        my $subnets = "";
        $subnets = join " ", map {"$_->{subnet}/$_->{network_bits}"} @ifs;
        return $subnets;
    }
    warn "Unable to calculate default for $d";
    return "";
}

sub write_config {
    my ($self, $update) = @_;
    my %opts;
    %opts = %{$self->{inst_opts}} if exists $self->{inst_opts};
    
    # Encrypt the manager password before writing it to the config file
    if (exists $opts{manager_password} and !$self->{_mgr_pass_crypted}) {
        $opts{webui_password} = sha1_hex($opts{manager_password});
        $opts{manager_password} = mk_password($opts{manager_password});
        inst_log("Encrypted manager_password");
    }

    # don't write options that are invalid
    foreach my $o (keys %opts) {
        my $options = $self->{iq_conf}{option};
        next unless $options and ref($options->{$o}{must_match}) eq "ARRAY";
        inst_log("Dropping invalid option $o from install.conf");
        delete $opts{$o} unless 
            grep {$_ eq $opts{$o}} @{$self->{iq_conf}{$o}{must_match}};
    }

    my @pkgs = @{$self->get_packages};
    if($update || $self->{auto_config}) {
        my $conf = PureMessage::Install::Config::read_config(
                        $update ? undef : $self->{auto_config}
        );
        push @pkgs, @{$conf->{pkgs}};
        foreach my $k (keys %{$conf->{inst_opts}}) {
            $opts{$k} ||= $conf->{inst_opts}{$k};
        }
    }

    # determine our roles by calculating from installed packages,
    # and from what we are installing now.
    my %roles = installed_roles($self->{prefix}, $self->{repo}, @pkgs);
    $roles{$_}++ foreach @{$self->{roles}};

    my $user = $opts{pmx_user} or scalar getpwuid((stat("$self->{prefix}/bin"))[4]);
    my $fn = PureMessage::Install::Config::write_config( 
                                                     pkgs => \@pkgs,
                                                     opts => \%opts,
                                                     pmxuser => $user,
                                                     roles => [ keys %roles ],
                                                   );
    inst_log("Wrote install config file to $fn");
}

sub load_config {
    my ($self, %subopts) = @_;

    my $rios_only = delete $subopts{rios_only};
    my $conf;
    eval {
        $conf = PureMessage::Install::Config::read_config(delete $subopts{config});
    };
    if ($@) {
        inst_log("Failed to read install config: $@");
        return;
    }
    $self->{pkgs} = $conf->{pkgs} unless $rios_only;
    inst_log("Read install config file");

    # Go through the RIOS, and make sure they are all answered
    my $rio = $self->get_first_rio();
    while(defined $rio) {
        my $name = $rio->{name};
        inst_log("load_config: loading option $name");
        if (defined $conf->{inst_opts}{$name}) {
            $self->{_mgr_pass_crypted}++ if ($name eq "manager_password");
            eval {
                my $q = $self->set_instopt($name, $conf->{inst_opts}{$name}, 
                                           %subopts);
                if (ref($q) eq "HASHREF") {
                    die "Install option $name requires suboption $q->{subopt_name}";
                }
            };
            if ($@) {
                inst_log($@);
            }
            else {
                $rio->{prompt_user} = 0;
                inst_log("Required install option $name is valid");
            }
        }
        $rio = $self->get_next_rio();
    }
    inst_log("Successfully loaded install configuration");
}

sub create_pmx_user {
    my $self = shift;

    my $user = $self->{inst_opts}{pmx_user};
    my $home = $self->{inst_opts}{pmx_home};
    unless(getpwnam($user)) {
        _create_user($user, $self->{inst_opts}{pmx_password}, $home);
    }
    # $home is different than $pw[7]
    unless (-d $home) {
        my @pw = getpwnam($user);
	if ($home eq $pw[7]) {
	    # gone missing for some reason
            mkdir $home or die "Can't mkdir $home: $!";
            chown @pw[2,3], $home or die "Can't chown $pw[2]:$pw[3] $home: $!";
	}
	else {
	    run("ln -s '$pw[7]' '$home' $logstr");
	}
    }

    my @pw = getpwnam($user);
    my $prefix = $self->{prefix} || die "Prefix is not defined!";

    my $extra_man = "$prefix/man:$prefix/sendmail/man:$prefix/postfix/man";
    if($^O !~ /^(linux|freebsd)$/) {
        $extra_man .= ":/usr/share/man";
    }
    
    inst_log("Installing PureMessage user shell environment in $pw[7]");

    # write_all_shell_files will also warn() with a more useful error message
    write_all_shell_files($pw[7], $pw[2], $pw[3], {
        prefix          => $prefix,
        ldpath          => $Config{ldlibpthname} || "LD_LIBRARY_PATH",
        version         => "5.200",
        extra_man       => $extra_man,
        http_proxy      => $ENV{http_proxy} || $ENV{HTTP_proxy} || ""
    }) or die "Failed to write shell files!";
    
    my $group = getgrgid($pw[3]) || $pw[3];
    run("\@chgrp $group '$prefix' $logstr");

    # add to cron.allow
    cron_allow_user($user);
    
    my $hushlogin = "$pw[7]/.hushlogin";
    unless (-f $hushlogin) {
	# touch .hushlogin to avoid noise in startup scripts
	if (open my $FH, ">", $hushlogin) {
	    close $FH;
            chown($pw[2], $pw[3], $hushlogin)
                or warn "Failed to change ownership of $hushlogin: $!";
	}
	else {
	    warn "Can't create '$hushlogin': $!";
	}
    }

    for my $d (qw(etc bin lib man doc html var)) {
	my $dir = "$prefix/$d";
        # don't re-chown these directories, as it will loose perms on files
        # already installed for a re-install case
        if (-d $dir) {
            my @stat = stat $dir;
            next if $stat[4] == $pw[2] and $stat[5] == $pw[3];
        }
	run("mkdir '$dir'") unless -d $dir;
	if ($d eq 'var') {
	    for my $d (qw(log qdir run)) {
		my $dir = "$prefix/var/$d";
		run("mkdir '$dir'") unless -d $dir;
	    }
	    run("chmod -R g+w $dir");
            run("chmod go-rwx $dir/qdir");
	}
        elsif ($d eq 'lib') {
            my $dir = "$prefix/lib/manager";
            run("mkdir '$dir'") unless -d $dir;
        }
	run("chown -R $user:$group $dir");
    }
    #

    inst_log("The user '$user' has now been configured.");
    my $pmx_sh = $self->{pmx_sh};
    if ($pmx_sh and -e $pmx_sh) {
        (my $basename = $pmx_sh) =~ s,.*/,,;
        my $h_sh = "$pw[7]/$basename";
        run("cp", $pmx_sh,  $h_sh) unless $h_sh eq $pmx_sh;
        run("chown", "$user:$group", $h_sh);
        run("chmod", "+x", $h_sh);
        inst_log("Copied the shell installer to $h_sh");
        $self->{user_pmx_sh} = $h_sh;
    }

    # pmx user should own the install log too
    run("chown", "$user:$group", $self->{log_file});

    # If we are using a local repository, create a link for the pmx user.
    # This allows CD installs to continue.
    (my $repo = $self->{repo}) =~ s#^file:(.+)/$#$1#;
    inst_log("Checking for local repo: repo=($repo)");
    if (-d $repo && !-e "$self->{prefix}/home/pmxrepo") {
	my $target;
	if (-l $repo) {
	    $target = readlink($repo) or die "can't readlink $repo: $!";
	}
	else {
	    $target = $repo;
	}
	su(user => $user, login => 1, silent => 1,
	   command => "ln -s $target $self->{prefix}/home/pmxrepo $logstr");
        $self->{repo} = "file:$self->{prefix}/home/pmxrepo/";
        PureMessage::Install::Upgrade::set_repository($self->{repo});
    }
}

# Options:
# - user => user name to create
# - pass => user password to set
# - home => user home directory
sub _create_user {
    my ($user, $pass, $home) = @_;
    if (-d $home and (stat _)[4] == 0) {
	# novice admins try to create the install directory
	# as root before running the installer, so we explicitly
	# check for this case and disallow it.
        die "$home exists, and is owned by 'root'!  Please remove or rename "
            . "this directory and run this step again.";
    }

    my $group = $user;
    my $ran_groupadd;
    my $ran_useradd;
    eval {
        unless (defined(getgrnam($group))) {
            inst_log("Creating group $group");
            groupadd($group);
            $ran_groupadd++;
        }
        $group = [ $group ];

        # Add the user to the cron group, if it exists
        if (-f "/etc/gentoo-release" and getgrnam("cron")) {
            push @$group, "cron";
        }
        inst_log("Creating user=$user groups=(" . join(", ", @$group) 
                 . ") home=$home");
        useradd( user => $user,
                 comment => "PureMessage user",
                 group   => $group,
                 create_home => 1,
                 home => $home);
        $ran_useradd++;
    };
    if ($@) {
        userdel(user => $user, remove_home => 1) if $ran_useradd;
        groupdel(group => $group->[0], _nocroak => 1) if $ran_groupadd;
        die "Error creating PureMessage user: $@";
    }
    # Set the password of the pmx user
    passwd( user => $user, password => $pass );

    # Now check that the user exists
    my @pw = getpwnam($user);
    die "Can't locate '$user' in the password database" unless @pw;
}

sub install_license {
    my $self = shift;
    my $pmx_sh = $self->{user_pmx_sh} || 
                 $self->{pmx_sh} ||
                 die "No shell installer was specified!\n";
    unless(pmx_can_run($pmx_sh)) {
        die "$pmx_sh cannot be run by the PureMessage user.\n";
    }

    inst_log("Installing new license from $pmx_sh");
    run_as_pmx( prefix => $self->{prefix},
                command => "$0 --install-license $pmx_sh "
                           . ">> $self->{log_file} 2>&1");
}

sub install_perl {
    my $self = shift;
    my $tmpdir;
    my $pmx_sh = $self->{user_pmx_sh} || die "pmx_sh is not defined!";
    my $prefix = $self->{prefix};
    my $repo   = $self->{repo};
    my $perl   = "$prefix/bin/perl";
    my $ppm    = "$prefix/bin/ppm";

    if (!-x $perl || ($ENV{PMX_REINST_PERL})) {
        my $arch = $^O;
        # XXX temporarily disable prompting about solaris64 unless
        # an environment variable is defined. That way we can
        # still tell customers about it if they *really* want
        # 64-bit support, and we can still test it, but it doesn't
        # mess up regular customers.
        if ($^O eq "solaris" and $self->get_instopt("solaris64")) {
            my $isainfo = `isainfo 2>/dev/null`;
            if ($? or $isainfo !~ /sparcv9/) {
                die <<EOT;
This version of Solaris is not capable of running 64-bit applications. The 64-bit version of PureMessage can be installed to work around limitations in the 32-bit stdio library.

Should you choose to upgrade to the 64-bit version, you will have to reinstall PureMessage.
EOT
            }
            else {
                $arch .= "64";
            }
        }

        inst_log("Installing PureMessage-Perl for $arch");
        install_tarball($prefix, $repo, 'PureMessage-Perl', 'NO PPM PREREQS');

        run_as_pmx(prefix => $prefix,
                   command => qq("$ppm" set repository PureMessage "$repo" $logstr)
                );
        inst_log("Set PureMessage PPM repository to $repo");
        
        run_as_pmx(prefix => $prefix,
                   command => qq("$ppm" set force_install 0 $logstr)
                );
        inst_log("Set PPM to not continue when dependencies fail.");
    }

    # Now that ppm is installed, "register" pmx-setup and PM-Perl as installed
    _update_ppm_pkg( prefix => $prefix, 
                     repo   => $repo,
                     pkg    => "pmx-setup");

    _update_ppm_pkg( prefix => $prefix, 
                     repo   => $repo,
                     pkg    => "PureMessage-Perl");
}

# fixup the permissions and owenership and install test & real cert
sub fixup_validator {
    my $self = shift;
    my $prefix = shift || $self->{prefix};
    my @pw = pmx_user($prefix);
    my $pmx_uid = $pw[2] || die "Can't find user id for pmx";
    my $pmx_gid = $pw[3] || die "Can't find group id for pmx";

    chown $pmx_uid, $pmx_gid, "$prefix/bin/pmx-validator";
}

# Functions to perform special install actions for specific packages
my %package_installers = ( 'PureMessage-Sendmail' => \&install_tarball,
                           'PureMessage-Postfix'  => \&install_tarball,
                           'PureMessage-JSMS'     => \&install_tarball,
                           'PureMessage-Install-PrivTools'
                                                  => \&install_tarball,
                           'PureMessage-Apache'   => \&install_tarball,
                           'PureMessage-PostgreSQL' => \&install_tarball,
                           'PureMessage-Perl'     => \&install_tarball,
                           'pmx-setup'            => \&upgrade_pmx_setup,
                           'pmx-setup-v5'         => \&upgrade_pmx_setup,
                         );

sub install_pkg {
    my ($self, $pkg) = @_;

    my $ppm = "$self->{prefix}/bin/ppm";
    die "PureMessage-Perl is not yet installed!" unless -x $ppm;

    unless (defined $self->{pkgs_installed} and %{$self->{pkgs_installed}}) {
        open(my $query, "$ppm query|") || die "Can't run '$ppm query'";
        local $_;
        while (<$query>) {
            next unless /^([-\w]+)/;
            $self->{pkgs_installed}{$1}++;
        }
    }

    if ($self->{pkgs_installed}{$pkg}) {
        inst_log("Package $pkg is already installed!");
        return 0;
    }

    eval {
        my $msgs = install_package(
            prefix => $self->{prefix}, 
            repo => $self->{repo}, 
            pkg => $pkg);
        push @{$self->{pkg_msgs}}, @$msgs;
    };
    if ($@) {
        inst_log("Error installing package $pkg: $@");
        die "Error installing package $pkg";
    }
    $self->{pkgs_installed}{$pkg}++;
    return 1;
}

sub get_install_msgs {
    my $self = shift;
    inst_log("Returning " . @{$self->{pkg_msgs}} . " package messages");
    return $self->{pkg_msgs};
}

sub install_package {
    my %opts = @_;

    my $prefix = delete $opts{prefix} or die "Prefix is mandatory";
    my $repo = delete $opts{repo} or die "Repository is mandatory";
    my $pkg = delete $opts{pkg} or die "Package is mandatory";
    my $force = delete $opts{force} || "";

    # Monitor the install log for trouble!
    my $logfile = inst_log;
    my $logfh;
    unless(open($logfh, $logfile)) {
        inst_log("Can't open $logfile for reading: $!");
    }
    seek $logfh, 0, 2 if defined $logfh; # go to EOF

    my $verify_opts = get_no_verification() ? '--novalidate' : '';

    my $ppm = "$prefix/bin/ppm";
    inst_log("Installing package $pkg");
    if (exists $package_installers{$pkg}) {
        $package_installers{$pkg}->($prefix,$repo,$pkg,%opts);
    }
    else {
        my $inst_log = inst_log();
        if ($pkg eq "PPM") {
            # Older PPMs really are quite brittle. They can't use "file:/path"
            # URLs (they want plain old paths), and they don't understand the
            # new "verify --upgrade --install" option.
            (my $ppm_repo = $repo) =~ s/^file://;

            # Force install the package directly from the repo
            # this method works well even with old ppms regardless of 
            # wether the package is uninstalled, installed, or up-to-date
            run_as_pmx(prefix => $prefix,
                        # XXX: verify here as well?
                       command => "$ppm install ${ppm_repo}PPM.ppd $logstr");
        }
        else {
            # We shouldn't use the --install option when updating PPM because
            # older PPM's may not support it.
            run_as_pmx(prefix => $prefix,
                       command => "$ppm verify --upgrade $verify_opts $force --install "
                                  . "--location=$repo $pkg $logstr");
        }
    }
    inst_log("Installed package $pkg");

    # Now check the install log for emitted messages
    my $type;
    my @msg;
    my @pkg_msgs;
    my $push_msg_sub = sub {
        push @pkg_msgs, { type => $type,
                          msg => join("\n", @msg),
                        };
        $type = undef;
        @msg = ();
    };
    while (defined(my $l = <$logfh>)) {
        unless($l =~ /^(info|err):\s+(.+)$/i) {
            $push_msg_sub->() if defined $type and @msg;
            next;
        }
        my $newtype = lc $1;
        $push_msg_sub->() if defined $type and $type ne $newtype;

        $type = $newtype;
        push @msg, $2;
    }
    $push_msg_sub->() if defined $type and @msg;
    close $logfh;

    return \@pkg_msgs;
}

sub upgrade_pmx_setup {
    my ($prefix, $repo, $pkg, %opts) = @_;

    my $setup = "$prefix/bin/pmx-setup-v5";
    eval {
        _fetch("${repo}pmx-setup-v5-$^O", $setup, $repo);
        chmod 0755, $setup or die "Couldn't chmod 0755 $setup: $!";
    };
    if ($@) {
        inst_log("failure fetching pmx-setup-v5: $@");
        rename "$0.old", $0;
        die $@;
    }

    # now update ppm
    if (-x "$prefix/bin/ppm") {
        _update_ppm_pkg( prefix => $prefix, 
                         repo => $repo,
                         force => 1,
                         pkg => "pmx-setup");
    }

    # restart ourself
    if ($opts{restart}) {
        $opts{restart_args} ||= "";
        print "Restarting pmx-setup-v5 ...\n";
        inst_log("Successfully updated pmx-setup-v5. restarting ...");
        exec("$prefix/bin/pmx-setup-v5 $opts{restart_args}");
    }
}

sub _fetch {
    my ($from, $to, $repo) = @_;

    inst_log("Fetching $from -> sophos-validation-escrow");
    my $file = fetch_file($from, $repo);

    if (!$file) {
        die "Error verifying file '$from': $PPM::Validate::PPMERR";
    }

    inst_log("Doing atomic rename of file '$file' to '$to'");

    eval {

        # rename/rename/unlink instead of just rename since some UNIX
        # flavors don't allow deleting an executable while a process is
        # running from it (but it can be renamed)...

        unlink "$to.old" if -e "$to.old"; # remove and old ones that are there

        rename($to, "$to.old");
        rename($file, $to) or die "Can't rename $file to $to: $!";
        unlink "$to.old";

    };
    if ($@) {
        rename("$to.old",$to) if -e "$to.old";
        die "Error copying from '$file' to '$to': $@";
    }

}

sub install_tarball {
    my ($prefix, $repo, $pkg, $no_ppm_prereqs) = @_;
    my $tgz = "$pkg-$^O.tar$gz";
    my $tgz_url = "${repo}$tgz";

    # must be root to continue
    unless ($> == 0) {
        inst_log("Must be root to upgrade $pkg");
        die "Must be root to upgrade $pkg";
    }

    # install any prerequisites first
    _install_ppm_prereqs( prefix => $prefix, 
                          repo => $repo,
                          pkg => $pkg)
            unless $no_ppm_prereqs;

    my $tmpdir = create_tmp_dir($pkg);
    eval {
        inst_log("Retrieving $tgz_url to $tmpdir $logstr");
        _fetch($tgz_url, "$tmpdir/$tgz", $repo);
	run("cd $tmpdir && ($zcat $tgz | tar xf -) $logstr");

        unlink "$tmpdir/$tgz" or die "Can't delete $tmpdir/$tgz: $!";

    # Old packages all assume that install will be run as root.
    if (-e "$tmpdir/install") {
        # Run as root
	    run(qq($tmpdir/install "$prefix" $logstr));
    }

    # Going forward, anything to be done as root gets put in install_root
    if (-e "$tmpdir/install_root") {
	    run(qq($tmpdir/install_root "$prefix" $logstr));
    }

    # Install steps to be done as the pmx user:
    if (-e "$tmpdir/install_pmx") {
        run_as_pmx(prefix => $prefix,
                   command => "$tmpdir/install_pmx $prefix $logstr",
                   );
    }

	if (my @u = pmx_user($prefix)) {
	    # If the package is installed after a PureMessage
	    # user is created, but before any PureMessage packages has
	    # been installed, then we must ensure that $pkg_dir
	    # continues to be writeable by the PureMessage user.
	    my $pkg_dir = "$prefix/etc/.packages";
	    run("chown $u[2]:$u[3] '$pkg_dir' $logstr")
		unless (stat $pkg_dir)[4] == $u[2];
	}
    };
    my $err = $@;
    run("-rm", "-rf", $tmpdir);
    if ($err) {
        inst_log("Couldn't install $pkg tarball: $err");
        die "Couldn't install $pkg tarball";
    }

    _update_ppm_pkg( prefix => $prefix, 
                     repo => $repo,
                     pkg => $pkg);
}

sub _install_ppm_prereqs {
    my %opts = @_;
    my $ppm = "$opts{prefix}/bin/ppm";
    eval {
        run_as_pmx( prefix => $opts{prefix},
                    command => "$ppm verify --prereq-only "
                               . "--upgrade --install "
                               . "--location=$opts{repo} $opts{pkg} $logstr" );
    };
    if ($@) {
        inst_log("Couldn't ppm install prerequisites for $opts{pkg}: $@");
        die "Error during ppm install prerequisites for $opts{pkg}";
    }
}

sub _update_ppm_pkg {
    my %opts = @_;
    my $ppm = "$opts{prefix}/bin/ppm";
    $opts{force} = $opts{force} ? "--force " : "";
    eval {
        # this is tricky - we need to have PPM_JUST_DO_IT set when
        # running ppm to have it "install" this package.
        # run_as_pmx will use su() and pmx-env, so we need to sh it
        run_as_pmx( prefix => $opts{prefix},
                    command => "sh -c 'PPM_JUST_DO_IT=1 $ppm verify "
                               . "--upgrade --install $opts{force} "
                               . "--location=$opts{repo} $opts{pkg} $logstr'" );
    };
    if ($@) {
        inst_log("Couldn't ppm install $opts{pkg}: $@");
        die "Error during ppm install of $opts{pkg}";
    }
}

sub finalize_install {
    my ($self,$progress_cb) = @_;
    my $found_initd = 0;
    foreach my $d (qw(/etc/init.d /etc/rc.d/init.d)) {
        next unless -d $d;
        $found_initd++;
        if (!-l "$d/puremessage" and -e "$self->{prefix}/bin/pmx-init") {
            run("ln -s $self->{prefix}/bin/pmx-init $d/puremessage");
            inst_log(<<EOT);

    The PureMessage system startup program 'pmx-init' has now been
    linked to $d/puremessage. 

EOT
        }
    }
    unless ($found_initd) {
        inst_log(<<EOT);

    This system does not have an /etc/init.d directory so this
    setup program can not enable automatic startup of PureMessage
    mail filters by itself.  You will have to make sure that the
    command

        $self->{prefix}/pmx-init start

    is executed by 'root' during system startup.

EOT
    }

    # initialize the resources system
    $self->initialize_resources();
    
    run("date > $self->{prefix}/etc/.pmx_installed");

    # Remove temp directories, etc.
    # We now control this directory completely so wipe it out everytime
    my $temp_dir = tmp_dir_location();
    rmtree $_ for glob("$temp_dir/*");

    # create package upgrades statefile
    eval {
        PureMessage::Install::Upgrade::write_updates_info(0);
    };
    if ($@) {
        inst_log("Error creating package upgrades statefile: $@");
    }
    
    # delete the etc/.v5_fresh file
    # This file was potentially created by the v5 aware pmx-setup
    # to indicate a fresh install (and not a v4 upgrade). 
    my $fresh = "$self->{prefix}/etc/.v5_fresh";
    unlink $fresh if -e $fresh;

    # delete the etc/.v4_upgrade file, if present
    my $upgd_file = "$self->{prefix}/etc/.v4_upgrade";
    unlink $upgd_file if -e $upgd_file;
    
    unless (-f $NO_SNAPSHOT_FILE) {
        # make initial snapshot (delete it if it exists already)
        eval {
            my $ss = PureMessage::Snapshot->new( name => "Initial Install",
                                                 quiet => 1,
                                               );
            inst_log(q(Deleting existing "Initial Install" snapshot ...));
            $ss->delete if $ss;
        };
        if ($@ and $@ !~ /doesn't exist/) {
            inst_log("Error deleting existing initial snapshot: $@");
        }
        eval {
            PureMessage::Snapshot->new( name => "Initial Install",
                                        create => 1,
                                        desc => "Created after successful install",
                                        quiet => 1,
                                        progress_cb => $progress_cb,
                                      );
        };
        if ($@) {
            inst_log("Error creating initial snapshot: $@");
            die $@;
        }

        inst_log("Created \"Initial Install\" snapshot");
    }
    return;
}

sub initialize_resources {
    
    my $self = shift;
    my $resources_failed;
    eval {
	my $basecmd = qq|$self->{prefix}/bin/pmx-resources-init init --installer-errors|;
        run_as_pmx( prefix => $self->{prefix},
		    command => $basecmd . qq| --existing-dsn $logstr|,
                  );
        inst_log("Initialized the resource system for the default database host.");
    };
    if ($@) {
        $resources_failed++;
        inst_log(<<EOF);

Error initializing the resources system: $@;

########################################################
# There was an error initializing the resources system #
#                                                      #
# In order to use resources you will need to manually  #
# initialize using pmx-resources-init. Please refer to #
# the product documentation on the steps required.     #
#                                                      #
# Until you have initialized the resource system you   #
# will be unable to sychronize resources across        #
# hosts, view/generate reports or quarantine and       #
# manage email.                                        #
#                                                      #
########################################################


EOF
        eval {
            # we don't have a dsn so we'll just init the local store
            run_as_pmx( prefix => $self->{prefix},
                        command => "$self->{prefix}/bin/pmx-profile init --store=local $logstr"
                      );
            inst_log("Initialized the local resource store.");
        };
        
        if($@)
        {
            inst_log(<<EOF);

########################################################
# WARNING: The local store failed to be initialised!   #
#                                                      #
# The PureMessage Manager will not function until you  #
# run pmx-resources-init. Please refer to the product  #
# documentation on the steps required.                 #
#                                                      #
########################################################

EOF
        }
    }
    
    if ($resources_failed) {
        my $inst_log = inst_log;
        return <<EOT;
Failed to initialize resources.  Please see the install log at
($inst_log) for more information 
on how to fix this.
EOT
    }
}

sub reset_install {
    my $self = shift;
    $self->{pkgs_installed} = undef;
}

# This function is not called from the UI::Curses installer.
# The uninstall facility needs to be re-worked.  This code is all
# from pmx-setup, and not really thought through.
sub uninstall_milter {
    my ($prefix, $logfile) = @_;
    run("-$prefix/bin/pmx stop $logstr");

    # remove the rest of the Manager plug-in's
    foreach my $mdir (qw( etc/manager lib/manager)){
        run("rm -rf $prefix/$mdir $logstr") if( -e "$prefix/$mdir" );
    }
    run("rm -f $prefix/etc/init.d/pmx-manager $logstr");

    my $pkg_dir = "$prefix/etc/.packages";
    opendir(my $d, $pkg_dir) || die "Can't open $pkg_dir: $!";
    my @pkg = grep /^PureMessage/ && $_ ne "PureMessage-Sendmail", readdir($d);
    closedir($d);
    run("cd '$pkg_dir' && rm -f @pkg $logstr");

    # delete pmx-* programs in the bin directory
    opendir($d, "$prefix/bin") || die "Can't open '$prefix/bin': $!";
    local $_;
    while (defined($_ = readdir($d))) {
        next unless /^pmx/;
        next if /^pmx-(setup|wfetch)$/;
        unlink("$prefix/bin/$_") || die "Can't unlink '$prefix/bin/$_': $!";
    }
    
    # final cleanups
    run("cd '$prefix/doc' && rm -rf pmx-* samples $logstr");
    run("find '$prefix/lib' -name PureMessage.pm -print | xargs rm -f $logstr");
    run("find '$prefix/lib' -type d -a -name PureMessage -print | xargs rm -rf $logstr");
}

sub uninstall_pmx_user {
    my $prefix = shift;
    my @u = pmx_user($prefix);
    return unless @u;
    my $uname = $u[0];
    my $home  = $u[7];

    return if $uname eq "root"; # safety guard
    cron_remove_user($uname);
    run("rm -f /var/spool/mail/$uname");
    userdel(user => $uname, remove_home => 1, _nocroak => 1);
    my $group = getgrgid($u[3]);
    groupdel(group => $group, _nocroak => 1) if $group && $group eq $uname;
    unlink("$prefix/home");
}

sub remove_prefix {
    my $prefix = shift;
    my @u = pmx_user($prefix);
    my $delayed_rm;
    if ($^O eq "hpux") {
        # On HP-UX we are not able to delete a program that
        # is currently running (which pmx-setup will be in this case)
        # but we are able to move the file.  If we don't move the
        # file away, them 'rm -rf' fails because the directory can't
        # be emptied.
        my $prefix_parent = $prefix;
        $prefix_parent =~ s,/[^/]+$,,;
        $delayed_rm = "$prefix_parent/pmx-setup-$$";
        run("-mv $prefix/bin/pmx-setup-v5 $delayed_rm $logstr");
    }
    chdir("/"); # avoid warning if we happen to delete current directory
    $logstr = "> /dev/null 2>&1";
    my $v4 = -e "$prefix/etc/.version-pmx-4";
    
    if (@u) {
	foreach my $folder (<$prefix/*>) {
	    chomp $folder;
	    run("-rm -rf $prefix/$folder $logstr")
		unless "$prefix/$folder" eq $u[7];
	}
    }
    else {
	run("-rm -rf $prefix $logstr");
    }
    
    run("-rm -f /etc/puremessage-5 $logstr"); 
    run("-rm -f /etc/puremessage-4 $logstr") if $v4;
    run("-(sleep 3; rm -f $delayed_rm) $logstr &") if $delayed_rm;
}

sub eula_accepted {
    my ($prefix, $accepted) = @_;

    my $eula_file = "$prefix/etc/.eula-accepted";
    if (!defined($accepted)) {
        # return eula status
        return -f $eula_file;
    }

    inst_log("EULA was accepted");
    unlink($eula_file);
    open(my $x, ">", $eula_file) ||
        die "Can't create $eula_file: $!";
    print $x scalar localtime, "\n";
    close($x) || die "Can't write to $eula_file: $!";

    return 1;
}

sub installed_roles {
    my ($prefix, $repo, @pretend_pkgs) = @_;
    die "No repository specified!" unless defined $repo;
    my $iq = get_iq_conf($repo);
    my @roles;
    my %pretend = map {$_ => 1} @pretend_pkgs;

    foreach my $r (keys %{$iq->{role}}) {
        next if $r eq "desc" or $r eq "default";
        my $installed;
        if (defined $iq->{role}{$r}{any_installed}) {
            foreach my $p (@{$iq->{role}{$r}{any_installed}}) {
                if (-e "$prefix/etc/.packages/$p" or $pretend{$p}) {
                    $installed++;
                    inst_log("Role $r is installed because of package $p");
                    last;
                }
            }
        }
        else {
            # Check the packages that are associated with this role.  There 
            # are some packages that don't use AS::Install, and so don't have
            # files in .packages.  We can ignore these packages (assume they
            # are installed).
            my %ignore_pkgs = ( "ActiveState-Milter" => 1,
                              );
            # all packages must be installed
            $installed++;
            foreach my $p (@{$iq->{role}{$r}{packages}}) {
                unless (-e "$prefix/etc/.packages/$p" or $ignore_pkgs{$p}
                        or $pretend{$p}) {
                    $installed = 0;
                    inst_log("Role $r is not installed because of package $p");
                }
            }
        }
        if ($installed) {
            push @roles, $r;
            inst_log("Role $r is installed");
        }
    }

    return map {$_, 1} @roles if wantarray;
    return \@roles;
}

sub create_prefix_file {
    my $prefix = shift;
    my $pf = "/etc/puremessage-5";
    return if -e $pf and $prefix eq read_prefix_file($pf);
    
    inst_log("Appending $prefix to $pf");
    open(my $FH, ">>$pf") or die "Can't open $pf: $!";
    print $FH "$prefix\n";
    close $FH or die "Can't write $pf: $!";
}

1;

__END__

=head1 NAME

PureMessage::Install

=head1 SYNOPSIS

  package PureMessage::Install::Foo;
  use PureMessage::Install;
  our @ISA = ("PureMessage::Install");

=head1 DESCRIPTION

The PureMessage::Install package is a base class for different install types.

=head2 new()

The constructor uses the following arguments:

=over 4

=item B<pmx_sh>

The filename of the shell installer that invoked pmx-setup-v5.

=item B<prefix>

The PureMessage install prefix.

=item B<repo>

The PureMessage repository to retrieve packages from.

=back

=head2 calc_lic_comp( %opts )

Calculate the licensed components from the license file.  If perl is not
installed, then the list of packages returned shouldn't be considered 
"trustworthy".  One of the following options must be set:

=over 4

=item prefix

The current install prefix.

=item pmx_sh

The PureMessage shell installer script.

=head2 get_packages()

    # Get the list of packages to be installed
    my $pkgs = $inst->get_packages();
    foreach my $p (@$pkgs) { ... }

get_packages() returns an arrayref to the list of packages to be installed by 
the current installation type.

=head2 get_first_rio()
=head2 get_next_rio()
=head2 get_prev_rio()

Will return the first/next/previous Required Install Option (RIO) to be 
asked of the user.

Each function will return undef if no RIOs are available (all done or past 
first), or a hash containing the following keys:

=over 4

=item B<option_name>

Used to identify the RIO

=item B<default_value>

Default value to be displayed to the user

=item B<option_desc>

A textual description of the option.  This will be displayed to the user when
they are answering the question.

=item B<must_prompt>

If this is false, the default value can be used without prompting.

=back

=head2 get_rio_list()

Return the list of options and values from the user input.  An array of hashes
is returned.  Each hash contains "name" and "value" keys.

=head2 set_instopt( $opt, $val, [%subopts] )

Set the install option $opt to $val.  The install options will be written to 
the install.conf file.

Each install option may have some "sub-options".  If $val requires additional
information or decisions by the user, this function will return a hashref
containing a question to be posed to the user.  The hash will contain:

=over 4

=item B<subopt_name>

An identifier for this sub-option.

=item B<question>

A yes/no question to pose to the user.

=back

When the user answers the question, this function should be called again,
with the sub-option this time.

This function will throw an error if the specified value is not valid.
The error messages are suitable for presentation to the user.

=head2 get_instopt( $opt, $val )

Return the specified install option.

=head2 write_config()

Create the PureMessage install configuration file.  This file will be used by 
PureMessage packages during installation.

=head2 create_pmx_user()

Create the PureMessage user.  This function will raise errors if anything goes
wrong.

=head2 install_license()

Install the PureMessage license.  This will preform the license install as 
the PureMessage user.

=head2 install_perl()

Install PureMessage Perl.  This will perform the Perl install as the 
PureMessage user.

=head2 install_pkg( $pkg )

Install the specified package.  This will install the specified package as
PureMessage user.

=head2 finalize_install()

Perform any final actions after the install has finished successfully.  
Returns undef on success or a user friendly error message on error.

=head2 reset_install()

Reset any state about the current install process. 

=head2 get_install_msgs()

Return an arrayref of messages emitted during package installations.  Each 
element in the array is a hashref, containing keys for "type" and "msg".
Type will either be "info" or "err".

=head1 Class Functions

=head2 eula_accepted( $prefix, $accepted )

Set or retrieve the status of the EULA acceptance.  If $accepted is undef, the
current status will be returned.

=head2 install_package( prefix => $prefix, repo => $repo, pkg => $pkg )

Installs any PureMessage package.  Can be called without creating an install 
object.  Returns an arrayref of messages emitted during install.

=head2 installed_roles( $prefix, $repo, @pretend_pkgs )

@pretend_pkgs can be used to provide a list of other packages that should be
assumed to be installed when calculating what roles are installed.

In scalar context, an array ref of the installed roles is returned.  In array 
context, a list of key/value pairs will be returned.  The key is the role name,
and the value is true.

=head1 CONFIGURATION

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.

=head1 SEE ALSO

pmx-setup-v5
