package PureMessage::Install::Config;
use ActiveState::Config::Simple qw(config_parse_file);
use PureMessage::Install::Prefix qw($PREFIX);
use Exporter;
our @ISA = ('Exporter');
our @EXPORT_OK = qw(write_config read_config get_instopt get_roles
                    $install_conf);
use strict;

my $conf;
our $install_conf = "$PREFIX/etc/install.conf";

sub write_config {
    my (%opts) = @_;
    my $pkgs = $opts{pkgs} || [];
    my $inst_opts = $opts{opts} || {};
    my $roles = $opts{roles} || [];

    if (!defined $opts{pmxuser}) {
        # try to figure it out
        my $uid = (stat "$PREFIX/home")[4];
        $opts{pmxuser} = getpwuid($uid) if $uid;
    }
    die "No pmxuser provided!" unless defined $opts{pmxuser};

    my $prefix = $opts{prefix} || $PREFIX;
    my $dir = "$prefix/etc";
    die "$dir is not created!" unless -d $dir;

    open(my $fh, ">$install_conf") or 
        die "Can't open $install_conf for writing: $!";
    my $date = localtime;
    print $fh "# PureMessage Installation Configuration - created at $date\n",
              "<pkgs>\n";

    # NOTE: The fix for DEF29717 relies on the fact that the method
    # PureMessage::Install::Config->get_instopt() is broken, and returns a
    # defined value for 'CSM' and 'Solaris64' even if they are not set.
    print $fh "    $_\n" for @$pkgs;

    print $fh "</pkgs>\n\n<inst_opts>\n";
    foreach my $o (keys %$inst_opts) {
        next if $o =~ /pmx_password/;
        (my $val = $inst_opts->{$o}) =~ s/\\/\\\\/g;
        print $fh "    $o = \"$val\"\n";
    }
    print $fh "</inst_opts>\n";

    print $fh "# These roles are the ones that WILL be installed\n"
              . "<roles>\n";
    print $fh "    $_\n" foreach @$roles;
    print $fh "</roles>\n";
    close $fh;

    # If this is called from the "Save Configuration File & Exit" option, the
    # user may not exist yet.
    if (my @pw = getpwnam($opts{pmxuser})) {
	chown($pw[2], $pw[3], $install_conf)
	    or warn "Failed to change ownership of $install_conf: $!";
    }
    else {
        warn "Can't chown $install_conf - pmx user ($opts{pmxuser}) doesn't exist!\n";
    }

    # The install.conf may contain password information, so it shouldn't
    # be world readable
    chmod 0600, $install_conf or warn "Failed to chmod 0600 $install_conf: $!";

    return $install_conf;
}

sub get_instopt {
    my $opt = shift;
    eval { $conf ||= read_config() };
    if ($@) {
        warn $@;
        return;
    }
    if (exists $conf->{inst_opts}{$opt}) {
        return $conf->{inst_opts}{$opt};
    }
    warn "Warning - $opt is not an install option in $install_conf";
    return;
}

sub get_roles {
    eval { $conf ||= read_config() };
    if ($@) {
        warn $@;
        return;
    }
    my %roles = map {$_ => 1} @{$conf->{roles}};
    return \%roles;
}

sub read_config {
    my $install_conf = shift || $install_conf;
    die "The Install config file doesn't exist! ($install_conf)" 
        unless -f $install_conf;
    my $conf = config_parse_file($install_conf);
    $conf->{pkgs} = [] unless ref($conf->{pkgs}) eq "ARRAY";
    $conf->{roles} = [] unless ref($conf->{roles}) eq "ARRAY";
    return $conf;
}

1;

__END__

=head1 NAME

PureMessage::Install::Config

=head1 SYNOPSIS

  use PureMessage::Install::Config qw(write_config get_instopt);
  my $c = write_config(pkgs  => \@pkgs,
                       opts =>  \%install_options,
                       pmxuser => $username );

  my $foo = get_instopt("foo");

=head1 DESCRIPTION

The PureMessage::Install::Config module can create install.corf
config files, and retrieve values from them.

=head2 B<write_config( %opts )>

Writes the config options to the install config file.

=over 4

=item B<pkgs>

The PureMessage packages that will be installed.

=item B<opts>

The install options for this installation.

=item B<pmxuser>

The username of the PureMessage user.

=back

=head2 B<read_config()>

Reads the config options from the install config file.

=head2 B<get_instopt( $name )>

Return the value of the given install option.  An error will be raised if the
option doesn't exist in the install.conf file.

=head2 B<get_roles()>

Return a hashref containing the installed roles.  The keys of the hashref will
be the role name.  An error will be raised if the option doesn't exist in 
the install.conf file.

=head1 CONFIGURATION

The install config file is $PREFIX/etc/install.conf.

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.

=head1 SEE ALSO

PureMessage::Install
