package PureMessage::Install::OptionValidator;
use strict;
use Exporter;
our @ISA = ('Exporter');
our @EXPORT = qw(valid);
our @EXPORT_OK = qw(valid_email valid_domain valid_smtp valid_ips valid_user);
use Carp qw(croak);

sub valid {
    my ($type, $value) = @_;
    my $sub = { email =>  \&valid_email,
                domain => \&valid_domain,
                smtp =>   \&valid_smtp,
                ips =>    \&valid_ips,
                user =>   \&valid_user,
              }->{$type};
    croak "$type does not have a validator!" unless ($sub);
    return $sub->($value);
}

my $RE_hostname_part = qr/[0-9a-zA-Z][\w-]*/;
my $RE_hostname = qr/$RE_hostname_part(?:\.$RE_hostname_part)+/;
my $RE_ip_addr = qr/\d{1,3}(?:\.\d{1,3}){3}/;

sub valid_email {
    my $email = shift;
    return $email if ($email =~ /^[^\s@]+\@(?:$RE_hostname|$RE_hostname_part)$/);
    croak "$email is not a valid email address\n";
}

sub valid_domain {
    my $domain = shift;
    return $domain if $domain =~ /^$RE_hostname$/;
    croak "$domain is not a valid and fully qualified hostname\n";
}

sub valid_smtp {
    my $smtp = shift;
    if ($smtp =~ /^(smtp:)?($RE_hostname|$RE_hostname_part|$RE_ip_addr)(:\d+)?$/) {
        # try to fixup the URL
        $smtp = "smtp:" . $smtp unless $1;
        $smtp .= ":25" unless $smtp =~ /:\d+$/;
    }
    else {
        croak "$smtp is not a valid mail server\n";
    }
    return $smtp;
}

sub valid_ips {
    my $ips = shift;
    foreach (split ' ', $ips) {
        unless (/^$RE_ip_addr(?:\/\d+)?$/) {
            croak "$_ is not a valid IP address\n";
        }
    }
    return $ips;
}

sub valid_user {
    my $user = shift;

    unless ($user =~ /^[a-z][a-z0-9]{0,7}$/) {
        croak "The user name '$user' is not legal.  It can only contain "
            . "lowercase letters or numbers, and cannot be more than "
            . "eight characters long.\n";
    }
    if ($user =~ /^root|bin|adm|lp|sync|daemon|mail|news|ftp$/) {
        croak "The user name '$user' is reserved.\n";
    }
    return $user;
}

1;

__END__

=head1 NAME

PureMessage::Install::OptionValidator

=head1 SYNOPSIS

  use PureMessage::Install::OptionValidator;

  # check for a valid email address
  unless (valid('email', $email)) {
      die "Bad email address: $!";
  }

=head1 DESCRIPTION

This package stores commonly used validation routines for various PureMessage
configuration options.

=head2 valid( $type, $value )

Determine the validity of the given value.  The validated value is returned on
success, and an error is raised on failure.

$type can be one of the following:

=over 4

=item email

$value must be a valid email address.

=item domain

$value must be a valid and fully qualified hostname.

=item smtp_server

$value must be a valid smtp server.  The server is not contacted.

=item ips

$value must be a valid list of space seperated IP addresses.

=back

=head1 CONFIGURATION

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.
