package PureMessage::Install::Platform;

use strict;
use warnings;
use Exporter;
our @ISA = ('Exporter');
our @EXPORT_OK = qw(override supported $platform);

our $platform = "";

my %settings;
my $OVERRIDE;

sub override {
    $OVERRIDE = 1;
    %settings = @_;
}

sub supported {
    my $supported = 0;
    
    $settings{os} ||= $^O;
    chomp($settings{uname_m} ||= `uname -m` || "unknown");
    chomp($settings{uname_r} ||= `uname -r` || "unknown");
    chomp($settings{uname_v} ||= `uname -v` || "unknown");
    chomp($settings{uname_s} ||= `uname -s` || "unknown");
   
    if ($settings{os} eq "linux") {
	    find_distro() unless $OVERRIDE;
	    if (is_x86() || is_x86_64()) {
	        $supported = supported_linux();
	    } else {
	        $platform = "$settings{os} on $settings{uname_m}";
	    }
    }
    elsif ($settings{os} eq "freebsd") {
	    if (is_x86()) {
	        $supported = supported_freebsd();
	    }
	    else {
	        $platform = "$settings{os} on $settings{uname_m}";
	    }
    }
    elsif ($settings{os} eq "solaris") {
	    $supported = supported_solaris();
    }
    else {
	    $platform = "This operating system";
    }
    
    $platform ||= "This unknown version of $^O";
    
    return $supported;
}

## OS Subroutines ##

sub supported_linux {
    my $supported = 0;
    
    # Redhat
    if ($settings{redhat}) {
        # Sophos supports RedHat Enterprise 4, 5 (AS and ES)
	    if ($settings{redhat} =~ /^(?:4|5)(?:ES|AS)$/ || $settings{redhat} =~ /^5Server$/) {
	        $supported = 1;
	    }
	    $platform = $settings{redhat_friendly};
    }

    # Debian
    if ($settings{debian}) {
        # Sophos supports debian 4.0 and 5
	if ($settings{debian} =~ /^(\d+)/) {
            if ($1 == 4  || $1 == 5) {
		$supported = 1;
            }
	}
	$platform = "Debian Linux version $settings{debian}";
    }

    # SUSE
    if (exists $settings{suse_friendly} and 
        $settings{suse_friendly} =~ /Enterprise Server/) {
        # Sophos supports SuSE Enterprise Server 10 and 11
        # XXX: 
	    if ($settings{suse} =~ /^(?:10|11)$/) {
	        $supported = 1;
	    }
	    $platform = $settings{suse_friendly};
    }

    # Ubuntu
    if ($settings{debian} and $settings{uname_v} =~ /Ubuntu/) {
        $supported = 1;
        $platform = "Ubuntu Linux";
    }

    # Make sure we print a valid error message if we can't find out
    # the distribution
    $platform ||= "This distribution of Linux";

    return $supported;
}

sub supported_freebsd {
    my $supported = 0;
    # Sophos supports FreeBSD 6.3
    if ($settings{uname_r} =~ /^(\d+)\.(\d+)\.?\d*/) {
	    if (($1 == 6 || $1 == 7 ) and $2 == 3) {
	        $supported = 1;
	    }
	    $platform = "FreeBSD $settings{uname_r}";
    }
    return $supported;
}

sub supported_solaris {
    my $supported = 0;

    # In order to ensure that this is a supported version of Solaris,
    # we need to make sure that the SunOS version is 5.x (meaning that
    # the Solaris version is 2.x
    my ($sunos) = $settings{uname_r} =~ /^(\d)\./;
    return 0 unless ($settings{uname_s} eq "SunOS" and $sunos);

    # Check whether the machine hardware name is in the sun4 class
    if ($settings{uname_m} =~ /^sun4.$/) {
        if ($sunos == 5) { # Solaris 2.x
            # convert from a SunOS version to a solaris version
            $settings{uname_r} =~ s/[^.]*\.//; # Strip off the 5
            $platform = "Solaris $settings{uname_r}";
            
            if ($settings{uname_r} =~ /(\d+)/) {
                # Sophos supports Solaris 8 - 10
                if ($1 >= 9 and $1 <= 10) {
                    $supported = 1;
                }
            }
        }
        else {
            $platform = "SunOS$sunos";
        }
    }
    # Solaris x86
    elsif ($settings{uname_m} =~ /^i86pc$/) {
        if ($sunos == 5) { # Solaris 2.x
            # convert from a SunOS version to a solaris version
            $settings{uname_r} =~ s/[^.]*\.//; # Strip off the 5
            $platform = "Solaris $settings{uname_r}";
            
            if ($settings{uname_r} =~ /(\d+)/) {
                # Sophos supports Solaris 10
                if ($1 == 10) {
                    $supported = 1;
                }
            }
        }
    }
    else {
        # this is unsupported machine hardware
	    $platform = "Solaris on $settings{uname_m}";
	    return 0;
    }
    
    return $supported;
}

####

sub is_x86 {
    return $settings{uname_m} =~ /^i\d86$/;
}

sub is_x86_64 {
    return $settings{uname_m} =~ /^x86_64$/;
}

sub find_distro {
    # version info files for supported linux distributions:
    my $rh_cnf = "/etc/redhat-release";
    my $suse_cnf = "/etc/SuSE-release";
    my $deb_cnf = "/etc/debian_version";
    # SuSE Enterprise 8 doesn't have a SuSE-release file, instead
    # it has a UnitedLinux-release file
    my $united_cnf = "/etc/UnitedLinux-release";
    
    my $fh;
    if (-e $rh_cnf and not exists $settings{redhat}) {
	$settings{redhat} = `rpm -qf --qf "%{version}" $rh_cnf`;
	chomp($settings{redhat});
	unless (open($fh, "<$rh_cnf")) {
            warn "Can't open $rh_cnf\n";
            return 0;
        }
	chomp($settings{redhat_friendly} = <$fh>);
	close $fh;
        unless ($settings{redhat_friendly}) {
            warn "$rh_cnf is empty\n";
            return 0;
        }
    }
    if (-e $suse_cnf and not exists $settings{suse} ) {
	unless (open($fh, "<$suse_cnf")) {
            warn "Can't open $suse_cnf\n";
            return 0;
        }
	while (<$fh>) {
            chomp($settings{suse_friendly} ||= $_);
	    if (/VERSION = (.*)/) {
		chomp($settings{suse} = $1);
		last;
	    }
	}
	close $fh;
    }
    if (-e $deb_cnf and not exists $settings{debian}) {
	$settings{os} = "Debian Linux";
	unless (open($fh, "<$deb_cnf")) {
            warn "Can't open $deb_cnf\n";
            return 0;
        }
	chomp($settings{debian} = <$fh>);
	close $fh;
    }
    if (-e $united_cnf and not exists $settings{united}) {
        $settings{os} = "UnitedLinux";
	unless (open($fh, "<$united_cnf")) {
            warn "Can't open $united_cnf\n";
            return 0;
        }
	while (<$fh>) {
	    if (/VERSION = (.*)/) {
		chomp($settings{united} = $1);
		last;
	    }
	}
	close $fh;
    }
    
}

1;

__END__

=head1 NAME

PureMessage::Install::Platform

=head1 SYNOPSIS

  use PureMessage::Install::Platform qw( &supported $platform);

  # check for a supported platform
  my $supported = supported();
  
  if ($supported) {
     print "$platform is a supported platform\n";
  }
  else {
     print "$platform is not a supported platform\n";
  }

=head1 DESCRIPTION

This package is used to determine whether the platform being used is supported
in the current release of Sophos PureMessage.

=head2 supported()

This subroutine returns either 1 or 0 depending on whether the current platform 
is supported by Sophos PureMessage. If the platform cannot be determined,
either due to missing binaries or an unreadable version file, a warning will be
emitted and supported() will return 0.

supported() also sets the $platform variable to be a string version of the
platform. When the platform cannot be determined, $platform is set to $^O as a
last resort.

=head1 CONFIGURATION

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2005, Sophos Group. All Rights Reserved.
