package PureMessage::Install::Prerequisites;

use strict;
use warnings;

use base 'Exporter';

use PureMessage::_sys;
use PureMessage::Install::Prefix qw($PREFIX);
use PureMessage::Install::Log;

our @EXPORT_OK = qw(can_install
        prerequisite_error_msg
        can_upgrade
        upgrade_error_msg
        );

my $REASON = '';
my $UPGRADE_ERROR = '';

=item can_install()

    Usage:
        my $res = can_install(roles => ['database']);

    Returns whether or not the current system is able to install PureMessage.
    If the system does not meet the requirements to install, 0 is returned,
    and the reason for being unable to install can be accessed through
    prerequisite_error_msg().

=cut
sub can_install {
    my %opts = @_;
    my $roles = delete $opts{roles} or die "Must specify roles!";

    $REASON = '';
    # DB role requires a minimum shared memory and IPC settings
    if ( (grep /database/, @$roles)) {
        return 1 if -e '/etc/puremessage-no-db-mem-check';

        my %current = (
            SHMMAX => PureMessage::_sys::get_shmmax(),
            SHMALL => PureMessage::_sys::get_shmall(),
            SEMMNI => PureMessage::_sys::get_semmni(),
            SEMMNS => PureMessage::_sys::get_semmns(),
            );
        my %needed = (
            SHMMAX => 16 * 1024 * 103,
            SHMALL => 402,
            SEMMNI => 7,
            SEMMNS => 130,
            );
        my %descriptions = (
            SHMMAX => 'Maximum shared memory segment size',
            SHMALL => 'Maximum system shared memory',
            SEMMNI => 'Maximum number of semaphore sets',
            SEMMNS => 'Maximum number of semaphores (system-wide)',
            );

        my @not_adequate = ();
        foreach my $setting (keys %current) {
            my $cur = $current{$setting};
            next if ($cur < 0);  # Means that the setting wasn't defined for the platform
            if ($cur < $needed{$setting}) {
                push @not_adequate, $setting;
            }
        }

        if (@not_adequate) {
            $REASON = "The database role requires the following shared memory settings\n" .
                      "to be set properly:\n";

            foreach my $s (sort @not_adequate) {
                $REASON .= "$descriptions{$s} - must be at least $needed{$s} (currently set to $current{$s}).\n";
            }
            $REASON .= "Please consult the PostgreSQL tuning guide in the Sophos support\n" .
                       "knowledge base (available at http://www.sophos.com/support).";

            return 0;
        }

    } # DB role

    return 1;
}

sub prerequisite_error_msg { return $REASON; }

# returns -1 (version_a is smaller then version_b) , 0, or 1
sub _compare_versions {
    my $version_a = shift;
    my $version_b = shift;

    my @version_a = split('\.', $version_a);
    my @version_b = split('\.', $version_b);
    my $result = 0;
    while (@version_a or @version_b) {
        my $a = shift @version_a || 0;
        my $b = shift @version_b || 0;
        no warnings 'numeric';
        $result = int($a) <=> int($b);
        last if $result != 0;
    }
    return $result;
}

sub _valid_postgres {
    my $PMX_VERSION = '0.0.0';
    $PMX_VERSION = PureMessage::Version::Version() if( eval { require PureMessage::Version; });

    inst_log("_valid_postgres()...$PMX_VERSION");

    # for pmx_version < 5.5.0 postgres cannot be symlinked from $PREFIX/postgres
    return 1 unless _compare_versions($PMX_VERSION, '5.5.0') == -1;
    if( -l "$PREFIX/postgres") {
        $UPGRADE_ERROR .= "$PREFIX/postgres has been detected as a symlink,\n" 
         . "the installation cannot proceed until this issue is fixed.\n"
         . "Please contact support or reference the product documentation\n"
         . "for installation guidance.\n"
         . "Postgres tuning guidelines are available in the\n" 
         . "Sophos Knowledge Base (available at http://www.sophos.com/support).";
        return 0;
    }
     
    return 1; 
}

=item can_upgrade()

    Returns whether or not the current system is able to be upgraded. 
    If the system does not meet the requirements to upgrade, 0 is returned,
    and the reason for being unable to upgrade can be accessed through
    upgrade_error_msg().

=cut
sub can_upgrade {
    $UPGRADE_ERROR = '';
    # check that postgres is installed in the correct directory

    inst_log("can_upgrade()...");
    return 0 if(not _valid_postgres());

    return 1;
}

sub upgrade_error_msg { return $UPGRADE_ERROR; }
 

1;
