package PureMessage::Install::Upgrade;
use strict;
use ActiveState::Unix::Pw qw(su);
use ActiveState::Handy qw(run);
use ActiveState::Config::Simple qw(config_parse_file);
use PureMessage::Install::Prefix qw($PREFIX);
use PureMessage::Install::Utils qw(get_iq_conf run_as_pmx);
use PPM::PureMessage::Utils qw(pmx_user);
use PureMessage::Install::Log;
use PureMessage::Install qw(install_package calc_lic_comp installed_roles);
use PPM::Validate qw(get_no_verification);
use Exporter;
our @ISA = ('Exporter');
our @EXPORT = qw(get_pkg_list check_for_updates perform_updates 
                 get_repository set_repository get_updates_info
                 upgrade_postgresql extract_db_migration_tool
                 write_updates_info);

my $pmx_Config;
my $repo_file = "$PREFIX/etc/.default_repo";
my $ppm = "$PREFIX/bin/ppm";
my $state_file = "$PREFIX/var/run/.updates_exists";

1;

# prefix = PMX install prefix
# repo = PMX repository (optional)
#        - if set, add default packages from the IQ file in this repo
# force
#       - if set, force get_iq_conf to update the package list from the
#         current repo
sub get_pkg_list {
    my $prefix = shift || "$PREFIX";
    my $repo = shift || get_repository();
    my $force = shift || undef;

    my $pkgs_dir = "$prefix/etc/.packages";
    opendir my $D, $pkgs_dir or return {};
    my %pkgs = map { $_ => "Unknown" } grep { !/^\./ } readdir $D;
    closedir $D;
    $pkgs{PPM} = "Unknown";

    # Make sure packages new to v5 are included for upgrade scenarios
    $pkgs{'PureMessage-Install'} ||= "Unknown";
    $pkgs{'PureMessage-Install-PrivTools'} ||= "Unknown";
    $pkgs{'pmx-setup'} ||= "Unknown";
    $pkgs{'PPM'} ||= "Unknown";
    if ($repo) {
        my $c = get_iq_conf($repo, $force);
        my %roles = installed_roles($prefix, $repo);
        foreach my $r ('default', keys %roles) {
            if (ref($c->{role}{$r}{packages}) eq "ARRAY") {
                foreach my $p (@{$c->{role}{$r}{packages}}) {
                    $pkgs{$p} ||= "Unknown";
                }
            }
        }
    }

    # include all licensed components (incl potentially uninstalled ones)
    my $lic_pkgs = calc_lic_comp(prefix => $prefix);
    foreach my $p (keys %$lic_pkgs) {
        $pkgs{$p} ||= "Unknown";
    }

    return \%pkgs;
}

sub check_for_updates {
    my %opts = ( pkgs => undef,
                 repo => undef,
                 callback => undef,
                 @_,
               );
    my $pkgs = delete $opts{pkgs} ||
               get_pkg_list() ||
               die "package list is mandatory";

    my $repo = delete $opts{repo} || 
               get_repository() ||
               die "repo is mandatory";

    my $updates_exists = 0;
    my $check;
    foreach my $p (sort keys %$pkgs) {
        $check .= "$p ";
    }
    
    my @pmx = pmx_user($PREFIX);
    return unless @pmx;
    
    my $validate_opt = '';
    if (get_no_verification()) {
        $validate_opt = '--novalidate';
    }
    my $cmd;
    if ($pmx[2] == $>) {
        $cmd = "$ppm verify $validate_opt --location=$repo $check";
    }
    else {
        $cmd = su( _norun => 1, user => $pmx[0],
                   command => "$ppm verify $validate_opt --location=$repo $check");
    }
    
    open(my $ppm_fh, "$cmd 2>&1 < /dev/null |") or 
        die "Couldn't execute $ppm: $!";
    while(<$ppm_fh>) {
        if (/^Package '([\w-]+)' is up to date\./) {
            $pkgs->{$1} = "- ";
            $opts{callback}->("Checking", $1) if defined $opts{callback};
        }
        elsif(/An upgrade to package '([\w-]+)' is available\./) {
            $pkgs->{$1} = "Yes";
            $opts{callback}->("Checking", $1) if defined $opts{callback};
            $updates_exists = 1 unless ($1 eq 'PureMessage-Sophos-Data' or
                                        $1 eq 'PureMessage-AntiSpam-Data');
        }
        elsif(/Package '([\w-]+)' has not been installed by PPM/ ) {
            $pkgs->{$1} = "NEW";
            $opts{callback}->("Checking", $1) if defined $opts{callback};
            $updates_exists = 1 unless ($1 eq 'PureMessage-Sophos-Data' or
                                        $1 eq 'PureMessage-AntiSpam-Data');
        }
    }
    close $ppm_fh;
    # save state file info
    write_updates_info($updates_exists);
}

sub perform_updates {
    my %opts = ( pkgs => undef,
                 repo => undef,
                 force => 0,
                 callback => undef,
                 @_,
               );
    my $force = $opts{force} ? "--force " : "";
    
    my @pmx = pmx_user($PREFIX);
    return [] unless @pmx;

    # Special sort
    my %pkgs = map { $_ => 1 } @{$opts{pkgs}};
    my @rank = qw(PureMessage-Perl
                  PPM
                  PureMessage-Install
                  PureMessage-Scheduler
                  PureMessage-Apache
                  PureMessage-PostgreSQL
                 );

    my @sorted_pkgs;
    foreach (@rank) {
       if ($pkgs{$_}) {
           push(@sorted_pkgs, $_);
           delete $pkgs{$_};
       }
    }
    map { push(@sorted_pkgs, $_) } sort { $a cmp $b } keys %pkgs;

    my @updated;
    my @pkg_msgs;
    foreach my $pkg (@sorted_pkgs) {
        $opts{callback}->("Updating", $pkg) if defined $opts{callback};
        eval {
            my $msgs = install_package(
                prefix => $PREFIX, 
                repo => $opts{repo}, 
                pkg => $pkg,
                force => $force);
            push @pkg_msgs, @$msgs;
        };
        if ($@) {
            if ($@ =~ /root/) {
                push @updated, "$pkg - error: must be root to update";
            }
            else {
                push @updated, "$pkg - error updating package";
            }
        }
        else {
            push @updated, "$pkg - upgraded";
        }
    }

    # save state file info
    write_updates_info(0);
    
    inst_log("found " . @pkg_msgs . " messages");
    return \@updated, \@pkg_msgs;
}

sub extract_db_migration_tool {
    my $bound = "pmx-4x-db-migrate";
    die "No PerlApp::extract_bound_file" 
        unless defined &PerlApp::extract_bound_file;
    my $local = PerlApp::extract_bound_file($bound);
    unless(defined $local) {
        my $err = "Fatal error - $bound is not bound!";
        inst_log($err);
        die $err;
    }
    return $local;
}

sub upgrade_postgresql {
    my %args = @_;
    my $cb = $args{progress_cb} || sub { 1 };
    my $prefix = $args{prefix} or die "prefix is mandatory";
    my $instlog = inst_log;
    my $logstr = ">> $instlog 2>&1";
    my $pgdir = "$prefix/postgres";

    $cb->("Extracting database migration tool");
    my $tool = extract_db_migration_tool();

    my $pg_uid = (stat($pgdir))[4];
    if ($pg_uid and $pg_uid != (stat("$prefix/home"))[4]) {
        my $pguser = getpwuid($pg_uid);
        my $output = "$prefix/home/${pguser}_cron.bak";
        inst_log("Saving cronjobs of $pguser to $output");
        $cb->("Saving ${pguser}'s cronjobs to $output");
        eval {
            # this may not work on 3rd world platforms
            run("\@crontab -u $pguser -l > $output");
            my @pmx = pmx_user;
            chown @pmx[2,3], $output or warn "Can't chown $output: $!";
            # set the crontab to an empty file (crontab -r sucks on freebsd)
            my $tmpfile = "/tmp/.pmx-setup-v5.empty.$$";
            run("touch $tmpfile");
            run("\@crontab -u $pguser $tmpfile");
            unlink $tmpfile;
        };
        if ($@) {
            inst_log("Error while saving $pguser cronjobs: $@");
        }
    }

    my $pmx_db = "$prefix/bin/pmx-database";
    my $qt_cmd = run_as_pmx( prefix => $prefix, _norun => 1,
                             command => "$prefix/bin/pmx-config quarantine_type");
    chomp(my $qt = `$qt_cmd`);
    if (-x $pmx_db and defined $qt and $qt =~ "pmdb") {
        $cb->("Starting PostgreSQL");
        eval {
            run_as_pmx( prefix => $prefix,
                        command => "$prefix/bin/pmx-database start $logstr");
        };
        die "Couldn't start pmx-database - check the install log." if $@;

        $cb->("Exporting 4.x data");
        my $dumpfile = "$prefix/home/pmx-4x-export.dmp";
        eval {
            run_as_pmx( prefix => $prefix,
                        command => "$prefix/bin/perl $tool --export "
                                   . "$dumpfile > $instlog 2>&1" );
        };
        if ($@) {
            run_as_pmx( prefix => $prefix,
                        command => "$prefix/bin/pmx-database stop $logstr" );
            die "Couldn't export 4.x data - Check the install log.";
        }

        $cb->("Stopping PostgreSQL");
        eval {
            run_as_pmx( prefix => $prefix,
                        command => "$prefix/bin/pmx-database stop $logstr");
        };
        die "Couldn't stop pmx-database - check the install log." if $@;
    }

    eval {
        $cb->("Backing up existing PostgreSQL");
        unless (-d "$pgdir.v4") {
            unless (rename $pgdir, "$pgdir.v4") {
                my $errstr = "Can't rename $pgdir to $pgdir.v4: $!";
                inst_log($errstr);
                die $errstr;
            }
            inst_log("Renamed $pgdir to $pgdir.v4");
            # we need to create the postgres directory, as prefix is 
            # only writable as root
            unless (mkdir $pgdir, 0755) {
                my $errstr = "Can't mkdir $pgdir: $!";
                inst_log($errstr);
                die $errstr;
            }
            inst_log("Created $pgdir, 0755");
        }
        $cb->("Changing permissions on existing PostgreSQL");
        my ($uid, $gid) = (pmx_user($prefix))[2,3];
        foreach my $d ($pgdir, "$pgdir.v4") {
            run("chown -R $uid:$gid $d");
            inst_log("chown'd $d to " . getpwuid($uid) . ':' 
                     . getgrgid($gid));
        }
    };
    die "While backing up existing PostgreSQL configuration:\n  $@" if $@;
}

sub get_repository {
    return unless -e $repo_file;
    my $repo;
    # Repo should be the first line in the file
    open(my $fh, $repo_file) or die "Can't open $repo_file: $!";
    chomp($repo = <$fh>);
    close $fh or die "Can't close $repo_file: $!";
    return $repo;
}

sub set_repository {
    my $repo = shift or return;
    open(my $fh, ">$repo_file") or die "Can't open $repo_file: $!";
    print $fh "$repo\n";
    close $fh or die "Can't write $repo_file: $!";
    my @ids = (pmx_user($PREFIX))[2,3];
    if (@ids == 2) {
        chown @ids, $repo_file or die "Can't chown @ids $repo_file: $!";
    }
    inst_log("Set default repository to $repo in $repo_file");

    if (-x $ppm) {
        my $logstr = " >> " . inst_log() . " 2>&1";
        eval {
            run_as_pmx( prefix => $PREFIX,
                        command => qq|$ppm set repository PureMessage '$repo'|
                                   . $logstr,
                      );
        };
        if ($@) {
            inst_log("Failed to set ppm repository to $repo: $@");
        }
        else {
            inst_log("Set ppm repository to $repo");
        }
    }
}

sub write_updates_info {
    # will write to a state file whenever new updates are available
    # to PureMessage.
    # will always write date and time when the check for updates was
    # performed.
    # receives only 1 arg, an indicator if updates actually exist or
    # not.
    # doesn't return any arguments (void function)
    
    my $updates_exists = shift;

    my (undef,$c_min,$c_hour,$c_mday,$c_mon,$c_year) = localtime(time);
    my $yyyy_mm_dd = sprintf("%d-%02d-%02d", $c_year + 1900, $c_mon + 1, $c_mday);
    my $hh_mm = sprintf("%02d:%02d", $c_hour, $c_min);

    my $status = "checked_date = $yyyy_mm_dd\n";
    $status .= "checked_time = $hh_mm\n";
    
    # indicate update availability in the state file
    $status .= "available = " . ($updates_exists ? "yes" : "no") . "\n";
    
    # write date/time of check to the state file
    # initially write to a temp file to prevent race conditions
    my $temp_state_file = "$PREFIX/var/run/.tmp_updates.$$";
    open(my $sfh, ">$temp_state_file") || 
        die "Can't open updates statefile $temp_state_file: $!";
    print $sfh $status;
    close $sfh || die "Can't write updates to statefile $temp_state_file: $!";

    # rename temp file to the proper name
    rename($temp_state_file, $state_file);
}

sub get_updates_info {
    # will read the date and time of the latest check for updates
    # along with the info if updates exists or not.
    # will return a hash ref

    # read updates info from the state file
    my $updates_info;
    
    eval {
        $updates_info = config_parse_file($state_file, undef);
    };
    
    if ($@) {
        $updates_info = {};
        inst_log("Failed to parse state file $state_file: $@");
    }

    return $updates_info;
}
