package PureMessage::Snapshot;
use Exporter;
use strict;
use PureMessage::Snapshot::Flavour;
use PureMessage::Snapshot::Config;
use PureMessage::Snapshot::Details;
use PureMessage::Snapshot::Log;
use PureMessage::Install::Prefix qw($PREFIX);
use PureMessage::Install::Utils qw(pmx_running);
use PPM::PureMessage::Utils qw(pmx_user);
use File::Path;
use Errno qw(EPIPE);
use Cwd qw(abs_path);
use IPC::Run qw(start finish);
our @ISA = qw(Exporter);
our @EXPORT = qw(snapshots);

my $verbose = 0;
my $interactive = -t STDIN;
$SIG{PIPE} = 'IGNORE';

sub new {
    my ($proto, %opts) = @_;
    die "no name provided!" unless defined $opts{name};
    my $class = ref($proto) || $proto;
    $verbose ||= delete $opts{verbose};
    my $self = { %opts };
    bless $self, $class;

    $self->{diff} ||= 0;        # default is full snapshot
    $self->{create} ||= 0;
    $self->{filespec} ||= "default";
    $self->_load_config();
    $self->{flavour} ||= $self->{_default_flavour};
    $interactive = 0 if $self->{quiet};

    if (!-d $self->{_ss_base_dir}) {
        ss_log("Creating base snapshot directory: $self->{_ss_base_dir}");

        my $umask_orig = umask(0); 
        my $rv = mkdir($self->{_ss_base_dir}, 0700);
        umask $umask_orig;
        unless ($rv) { die "Couldn't create base snapshot directory: $!\n"; }

        my @ids = (pmx_user)[2,3];
        chown @ids, $self->{_ss_base_dir} or 
            die "Couldn't set ownership of base snapshot directory: $!\n";
    }

    # Convert the user supplied name to a directory name
    $self->{_dir_name} = $self->_to_dirname;
    
    if ($self->{create}) {
        # check for existing snapshots
        if (-d $self->{_dir_name} ) {
            my $err = "Snapshot $self->{name} already exists!";
            print "$err\n" if $interactive;
            die $err;
        }
        eval {
            $self->_create();
        };
        $self->_create_abort($@) if ($@); # will die with error if $@
    }

    unless (-d $self->{_dir_name}) {
        die "Error - snapshot \"$self->{name}\" doesn't exist:\n"
            . "Can't find directory $self->{_dir_name}";
    }

    return $self;
}

sub _load_config {
    my ($self, $file) = @_;

    my $c = PureMessage::Snapshot::Config->new($file);
    $self->{_default_flavour} = $c->default_flavour;
    $self->{_ss_base_dir} = $c->snapshot_dir;
    $self->{_ss_log} = $c->snapshot_log;
}

sub _create {
    my ($self) = @_;
    my $callback = $self->{progress_cb};

    if ($self->{diff} and !-d "$self->{_ss_base_dir}/current") {
        if (-l "$self->{_ss_base_dir}/current") {
            # Dangling symlink - remove it
            unlink "$self->{_ss_base_dir}/current";
        }
        # check that a previous snapshot exists
        ss_log("Can't create diff snapshot: No previous snapshot exists!");
        $self->{diff} = 0;
    }

    ss_log("Creating snapshot ($self->{name}) at $self->{_dir_name}") 
        unless $self->{pretend};

    # Make the snapshot directories
    my $data_dir = $self->{_dir_name} . "/data";
    mkpath($data_dir) or die "Couldn't create $data_dir: $!";

    my $fobj = PureMessage::Snapshot::Flavour->new(flavour => $self->{flavour});
    $self->{_filespec} = $fobj->filespec;
    $self->{_typespec} = $fobj->typespec;

    # Create snapshot-details.conf (this file will be re-saved by
    # the pmx-snapshot-inst script as the appropriate user)
    my $sd = PureMessage::Snapshot::Details->new( ss_dir => $self->{_dir_name},
                                                  verbose => $verbose,
                                                );
    my $details_file = $sd->write( name => $self->{name},
                                   desc => $self->{desc},
                                   diff => $self->{diff},
                                   flavour => $self->{flavour},
                                   filespec => $self->{_filespec},
                                   typespec => $self->{_typespec},
                                 );

    my @snap = ("$PREFIX/bin/pmx-snapshot-inst");
    push @snap, "-p" if $self->{pretend};
    push @snap, "-d" if $self->{diff};
    push @snap, qw(--progress create),
	$self->{_dir_name};
    my $h = start \@snap, '>pipe', \*OUT
                or $self->_create_abort("pmx-snapshot-inst returned $?");
    
    my $total = <OUT>;
    if ($total =~ /^total files: (\d+)$/) {
	print "  Found $1 files to snapshot.\n" if $interactive;
    }
    else {
        die "Couldn't read total snapshot files: ($total)\n";
    }
    $total = $1;

    my $cnt = 0;
    while(<OUT>) {
	chomp;
	my ($pos,$filename) = /(\d+) (.*)/; 
	my $percent = $pos / $total * 100;
	&$callback(int $percent,"Adding $filename") if $callback;
    }

    close OUT || $self->_create_abort("Error closing output pipe");
    $h->finish;
    ss_log("Files added to snapshot") unless $self->{pretend};

    my $ssbase = $self->{_ss_base_dir};
    print "Updating $ssbase/current symlink ...\n" if $verbose;
    # Move the existing one out of the way
    if (-l "$ssbase/current") {
        unlink("$ssbase/current.prev") if -l "$ssbase/current.prev";
        rename("$ssbase/current", 
               "$ssbase/current.prev") 
            or $self->_create_abort("Could not create $ssbase/current.prev:"
                                    . " $!");
    }
    # Create the new one
    my $abs_ss_dir = abs_path($self->{_dir_name});
    unless (symlink($abs_ss_dir, "$ssbase/current")) {
        my $err = $!;
        if (-l "$ssbase/current.prev") {
            rename("$ssbase/current.prev", "$ssbase/current") or
                warn "Couldn't rename $ssbase/current.prev to "
                     . "$ssbase/current: $!";
        }
        $self->_create_abort("Could not create new current snapshot link "
                             . "from $abs_ss_dir to $ssbase/current: $err");
    }
    ss_log("Symlink to current was updated to $self->{_dir_name}")
        unless $self->{pretend};
    print "\nThe snapshot was successfully created.\n" if $interactive;
}

sub _create_abort {
    my ($self, $msg) = @_;
    print "Error creating snapshot: $msg\n" if $interactive;
    ss_log("Error creating snapshot: $msg");

    rmtree($self->{_dir_name}, !"print", !"skip_deletes");
    $_[0] = undef;
    
    die "The snapshot was not created successfully.\n";
}

sub restore {
    my ($self, $force) = @_;
    my $callback = $self->{progress_cb};

    if (pmx_running()) {
        die "Error - Cannot restore a snapshot while PureMessage is running";
    }

    ss_log("Restoring snapshot $self->{name} from $self->{_dir_name}") 
        unless $self->{pretend};
    print "  Restoring snapshot $self->{name}\n"
        if $interactive;
    my $flav = PureMessage::Snapshot::Flavour->new(flavour => $self->{flavour});

    my @snap = ("$PREFIX/bin/pmx-snapshot-inst");
    push @snap, "--force" if $force;
    push @snap, "-p" if $self->{pretend};
    push @snap, "-d" if $self->{diff};
    push @snap, qw(--progress restore),
	$self->{_dir_name};
    
    my $h = start
	\@snap,
	    '>pipe', \*OUT
	or $self->_restore_abort("pmx-snapshot-inst returned $?");
   
    my $total = <OUT>;
    chomp $total;
    unless($total =~ s/total files: (\d+)/$1/) {
	$self->_restore_abort($total);
    }
    
    while(<OUT>) {
	chomp $_;
	$self->_restore_abort($_)
	    unless (my ($pos,$filename) = /(\d+) (.*)/);
	my $percent = ($pos + $total) / ($total * 2) * 100;
	&$callback(int $percent,"Adding $filename ...") if $callback;
    }

    close OUT || $self->_restore_abort("Error closing output pipe");

    $h->finish;
    
    # Get pmx-snapshot-inst to do the work
    if ($h->result) {
        if ($interactive) {
            print STDERR "rc = " . $h->result . "\n";
            print STDERR <<EOM;
There was a fatal error while restoring the
"$self->{name}" snapshot.

The snapshot was not restored successfully.

EOM
        }
        die "Error restoring $self->{name}.";
    }
    elsif($interactive) {
	print "\nThe snapshot was successfully restored.\n";
    }
}

sub _restore_abort {
    my ($self, $msg) = @_;
    print "Error restoring snapshot: $msg\n" if $interactive;
    ss_log("Error restoring snapshot: $msg");

    die "The snapshot was not restored successfully.\n";
}


sub delete {
    my ($self, %opts) = @_;

    ss_log("Deleting snapshot $self->{name} from $self->{_dir_name}");

    # Get pmx-snapshot-inst to do the work
    my $prog = "$PREFIX/bin/pmx-snapshot-inst";
    my @args = ($prog, "delete", $self->{_dir_name});
    system(@args) == 0 or die "Problem executing ", join(" ", @args);
    if ($? << 8) {
        if ($interactive) {
            print STDERR "$prog rc = " . $? << 8 . "\n";
            print STDERR <<EOM;
There was a fatal error while deleting the
"$self->{name}" snapshot.

The snapshot was not deleted successfully.

EOM
        }
        die "Error while deleting $self->{name}";
    }
    print "Snapshot $self->{name} deleted.\n" if $interactive;
    $self = undef;
}

sub snapshots {
    my @ss = ();
    my $c = PureMessage::Snapshot::Config->new();
    my $ss_base = $c->snapshot_dir;

    return @ss unless -d $ss_base;
    opendir(my $dir, $ss_base) or die "Can't opendir $ss_base: $!";
    while(defined(my $d = readdir($dir))) {
        next if $d eq "." or $d eq "..";
        $d = "$ss_base/$d";
        next if -l $d;
        next unless -d $d;
        eval {
            push @ss, PureMessage::Snapshot::Details->new( ss_dir => $d )->as_hashref;
        };
        if ($@) {
            ss_log("Error determining snapshot details for $d: $@", 1);
        }
    }
    closedir($dir);
    return @ss;
}

sub dump {
    my $self = shift;
    print "PureMessage::Snapshot name = $self->{name}\n";
    foreach my $k (qw(create diff filespec)) {
        print "\t$k\t=> $self->{$k}\n";
    }
    print "\n";
    print "\tSnapshot directory: $self->{_ss_base_dir}\n";
    print "\tSnapshot log file:  $self->{_ss_log}\n";
    foreach my $k (sort keys %{$self->{_filespecs}}) {
        print "\n\tFilespec $k:\n";
        foreach my $s (@{$self->{_filespecs}{$k}}) {
            print "\t    $s\n";
        }
    }
}

sub _to_dirname {
    my ($self) = @_;
    my $name = $self->{name} or die "Name is not defined!";
    $name =~ s/\s+/_/g;
    $name =~ s/[^\w-]//g;
    $name = $self->{_ss_base_dir} . "/" . $name;
    return $name;
}

# TODO
# 
# - changed file permissions are not restored when an error occurs and the 
#   changes must be rolled back

1;


__END__

=head1 NAME

PureMessage::Snapshot

=head1 SYNOPSIS

  use PureMessage::Snapshot;

  my @snapshot_list = PureMessage::Snapshot->snapshots;

  # Create a snapshot
  my $s = PureMessage::Snapshot->new( name => "Awesome Config",
                                      create => 1,
                                      diff => 1,
                                    );

  # Restore a snapshot
  my $s = PureMessage::Snapshot->new( name => $snapshots[0]->{name},
                                    );
  $s->restore;

=head1 DESCRIPTION

The PureMessage::Snapshot module is used to create and restore system 
snapshots.

=head2 Creating a Snapshot

  # Create a snapshot
  my $s = PureMessage::Snapshot->new( name => "Awesome Config",
                                      desc => "super rad config that works",
                                      create => 1,
                                      diff => 0,
                                    );
Options:

=over 4

=item B<name>

The title of the new snapshot.

=item B<desc>

A short description of the snapshot (optional).

=item B<create>

True to create a snapshot, false (default) to load an existing snapshot.

=item B<diff>

True to create a diff snapshot, false (default) to create a full snapshot.

=item B<filespec>

Optional parameter to specify which filespec to snapshot.
        
=back

=head2 Restoring a Snapshot

  # Restore a snapshot
  my $s = PureMessage::Snapshot->new( name => $snapshots[0]->{name} );
  $s->restore( $force );

Restore Options:

=over 4

=item B<force>

If true, will not prompt when restoring the snapshot.  Default is false.

=back

=head2 Micellaneous Snapshot Functions

=head3 Retrieving a list of Snapshots

  my @snapshot_list = PureMessage::Snapshot->snapshots;

This function will return a list of existing snapshots.  Each element of the
returned array is a PureMessage::Snapshot::Details object.

=head3 Deleting a Snapshot

  my $s = PureMessage::Snapshot->new( name => $snapshots[0]->{name} );
  $s->delete;

Snapshots can be deleted with the delete() function.  Deleting a snapshot 
removes it from the system permanently.  The snapshot object is also deleted.

=head1 CONFIGURATION

All of the configuration options can be set in the etc/snapshot.conf file.

=over 4

=item B<snapshot_dir>

Specifies where the snapshots will be stored.  Defaults to var/snapshot
This option is relative to the PureMessage prefix if it is a relative path.

=item B<snapshot_log>

Specifies where the snapshot log should go.  Defaults to snapshot_log.
This option is relative to the PureMessage log_dir if it is a relative path.

=back

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.

=head1 SEE ALSO

PureMessage::Snapshot::Config

PureMessage::Snapshot::Flavour
