package PureMessage::Snapshot::Details;
use File::Find;
use strict;

sub new {
    my ($proto, %opts) = @_;
    my $class = ref($proto) || $proto;
    die "No snapshot dir provided!" unless $opts{ss_dir};
    die "\"$opts{ss_dir}\" is not a valid directory!" unless -d $opts{ss_dir};

    my $self = { verbose => $opts{verbose},
                 ss_dir => $opts{ss_dir},
               };
    bless $self, $class;
    return $self;
}

sub write {
    my ($self, %opts) = @_;

    $opts{date} = scalar(localtime);
    my $details = "$self->{ss_dir}/snapshot-details.conf";
    unlink $details if -e $details;
    open(my $fh, ">$details") or die "Could not open $details: $!";
    foreach my $k (keys %opts) {
        if (ref($opts{$k}) eq "ARRAY") {
            print $fh "<$k>\n";
            print $fh "  $_\n" foreach @{$opts{$k}};
            print $fh "</$k>\n";
            next;
        }
        $opts{$k} ||= 0;
        print $fh "$k = $opts{$k}\n";
    }
    close $fh or die "Could not close $details: $!";
    return $details;
}

sub as_hashref {
    my ($self, $dir) = @_;

    # maybe we did this alroady.
    return $self->{as_hash} if exists $self->{as_hash};

    my $details = "$self->{ss_dir}/snapshot-details.conf";
    open(my $fh, $details) or die "Can't open $details: $!";
    my %h = ();
    while(<$fh>) {
        if (/^(\S+)\s+=\s+(.+)$/) {
            $h{$1} = $2;
        }
        if (/^<(\S+)>$/) {
            my $foo = $1;
            while (<$fh>) {
                last if /^<\/$foo>$/;
                chomp;
                s/^\s*//;
                push @{$h{$foo}}, $_;
            }
        }
    }
    close $fh || die "Can't close $details: $!";
    $h{dir_name} = $self->{ss_dir};
    unless ($h{desc}) {
        $h{desc} = "No description";
    }
    $self->{as_hash} = \%h;
    return $self->{as_hash};
}


# convert the filespec and typespec into the hash of files and options
sub map_specs {
    my ($self, $prefix) = @_;
    my $d = $self->as_hashref;
    my $filespec = $d->{filespec};
    my $typespec = $d->{typespec};

    # maybe we did this already.
    return $self->{_files} if exists $self->{_files};

    my %files = ();
    foreach my $f (@$filespec) {
        print "mapping filespec ($f) ...\n" if $self->{verbose};
        unless ($f =~ /^(\S+)\s*(.*)$/) {
            warn "Invalid filespec line: $f";
        }
        my $what = $1;
        $what = "$prefix/$1" unless $1 =~ /^\//;
        my @opts = split /\s+/, $2;


        if (-d $what or -f _) {
            my @found = ();
            find( sub { push @found, $File::Find::name },
                  $what);

            # Include or exclude the files we found
            foreach my $a (@found) {
                next if $a =~ /\.ss_save$/;
                next if $a =~ /\.ss_bak$/;
                # Strip the PREFIX
                $a =~ s/^$prefix\/?//;
                if (grep {$_ eq "exclude"} @opts) {
                    print "Excluding $a ...\n" if $self->{verbose};
                    delete $files{$a};
                }
                else {
                    $files{$a} = { filename => $a,
                                   opts => {},
                                 };
                    map { $files{$a}->{opts}{$_}++ } @opts;
                }
            }
        }
        else {
            print "\tUnknown $what\n" if $self->{verbose};
        }
    }

    # Apply the typespec options
    foreach my $t (@$typespec) {
        print "mapping typespec ($t) ...\n" if $self->{verbose};
        unless ($t =~ /^\/(.+)\/\s+((\w+\s*)*)$/) {
            die "Invalid typespec line: $t";
        }
        my $regex = $1;
        my @opts = split(/\s+/, $2);
        eval { "" =~ /$regex/ };
        if ($@) {
            warn "Invalid typespec regex: $regex - skipping";
            next;
        }
        print "\tChecking for matches: regex=($regex) opts=("
              . join(", ", @opts) . ")\n" if $self->{verbose};
        foreach my $f (keys %files) {
            if ($f =~ /$regex/) {
                if (grep /exclude/i, @opts) {
                    print "   $f matches - excluding ...\n" if $self->{verbose};
                    delete $files{$f};
                    next;
                }

                print "   $f matches - adding options ...\n" if $self->{verbose};
                map { $files{$a}->{opts}{$_}++ } @opts;
            }
        }
    }

    # Convert the options list to a hash for easy option lookups 
    if ($self->{verbose}) {
        print "Filespec results:\n";
        foreach my $k (sort keys %files) {
            print "    $k => " 
                  . (keys(%{$files{$k}->{opts}}) 
                          ? join(", ", keys %{$files{$k}->{opts}}) 
                          : "No options")
                  . "\n";
        }
    }

    $self->{_files} = \%files;
    return $self->{_files};
}


1;

__END__

=head1 NAME

PureMessage::Snapshot::Details

=head1 SYNOPSIS

  use PureMessage::Snapshot::Details;
  my $sd = PureMessage::Snapshot::Details->new( ss_dir => $dir );

  # Create a new details file
  $sd->write( %opts);

  # Retrieve snapshot details
  my $hashref = $sd->as_hashref();

  # Map the filespec and typespec to a set of files
  my $files = $sd->map_specs();

=head1 DESCRIPTION

The PureMessage::Snapshot::Details module encapsulates the details of a 
given snapshot.

=head2 Creating a new PureMessage::Snapshot::Details object

  my $sd = PureMessage::Snapshot::Details->new( ss_dir => $dir, 
                                                verbose => 1,
                                              );

Options:

=over 4

=item B<ss_dir>

The full path to the directory containing a snapshot.

=item B<verbose>

True to provide verbose output to stdout.

=back

=head2 Writing Snapshot Details 

  $sd->write( option1 => value1, ... );

This function will create the details file in the snapshot directory
for the specified options.

=head2 Retrieving Snapshot Details

  # Restore a snapshot
  my $hashref = $sd->as_hashref;

This function will return a hashref with all of the details for the 
specified snapshot.  The details are:

=over 4

=item name

=item dir_name

=item desc

=item diff

=item flavour

=item filespec

=item typespec

=item date 

=back

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.

=head1 SEE ALSO

PureMessage::Snapshot::Config

PureMessage::Snapshot::Flavour



