package PureMessage::Snapshot::Flavour;
use ActiveState::Config::Simple qw(config_parse_file);
use PureMessage::Install::Prefix qw($PREFIX);
use Exporter;
use strict;

my $flavour_conf = "$PREFIX/etc/data/snapshot/flavour.conf";

sub flavours {
    my $h = config_parse_file($flavour_conf);
    my @flavs = ();
    foreach my $f (keys %{$h->{flavour}}) {
        push @flavs, { name => $f,
                       desc => $h->{flavour}{$f}{desc},
                     };
    }
    return @flavs;
}


sub new {
    my ($proto, %opts) = @_;
    die "No flavour provided!" unless $opts{flavour};
    my $class = ref($proto) || $proto;
    my $self = { %opts };
    bless $self, $class;

    # parse etc/data/snapshot/flavour.conf
    my $hash = config_parse_file($flavour_conf);
    my $f = $self->{flavour};
    die "$opts{flavour} is not a valid snapshot flavour!"
        unless exists $hash->{flavour}{$f};
    $self->{_flav_hash} = $hash->{flavour}{$f};

    return $self;
}


sub name {
    my ($self) = @_;
    return $self->{flavour};
}

sub desc {
    my ($self) = @_;
    return $self->{_flav_hash}{desc};
}

sub filespec {
    my ($self) = @_;
    return $self->{_flav_hash}{filespec};
}

sub typespec {
    my ($self) = @_;
    return $self->{_flav_hash}{typespec};
}

1;


__END__

=head1 NAME

PureMessage::Snapshot::Flavour

=head1 SYNOPSIS

  use PureMessage::Snapshot::Flavour;

  my @flavours = PureMessage::Snapshot::Flavour->flavours;

  my $f = PureMessage::Snapshot::Flavour->new($flavour);

=head1 DESCRIPTION

The PureMessage::Snapshot::Flavour module provides an object interface
to the data contained in the etc/data/snapshot/flavours.conf file.

Each flavour is encapsulated in an object, and functions are provided to 
retrieve the information.

In addition, this module can provide the list of flavours available.


=head2 Class Functions:

  # Retrieve a list of available flavours
  my @f = PureMessage::Snapshot::Flavour->flavours;

This function will return an array of hashrefs.  Each hashref has the 
following keys:

=over 4

=item B<name>

The name of the flavour - this can be passed to the 
PureMessage::Snapshot->new() method to create a new snapshot.

=item B<desc>

A short description of the flavour.

=back

=head2 Object Functions:

The following functions return information about the snapshot flavour.

=head3 name

Returns the name of the flavour.

=head3 desc

Returns a description of the flavour.

=head3 filespec

Returns an arrayref containing the filespec lines.

=head3 typespec

Returns an arrayref containing the typespec lines.

=head1 CONFIGURATION

All of the configuration for Snapshot flavours is stored in the 
etc/data/snapshot/flavours.conf file.

Each flavour must define the following options:

=over 4

=item desc

A description of the flavour.

=item B<filespec>

Specifies the list of files to be included and excluded in the snapshot.  

The filespec is processed from the top down.  Each entry in the filespec can 
contain a file or directory.  If a directory is specified, all 
files and directories under that directory are included (or excluded) as well.  
Finally, options may be specified (seperated by spaces) at the end of the 
line.  

=item B<typespec>

Allows additional options to be set for files matching certain patterns.

The following typespecs are supported:

=over 4

=item B<regex>

  /some_regex/ <options>

The regex typespec will set the specified options for all files in the
filespec matching the given pattern.

=back

=back 

The following options are supported:

=over 4

=item B<static>

This option provides a guarantee that any process
that changes this file will create a new file and rename it
into place.  This option permits us to create a hard link
instead of copying the file, which can result in much better performance.

Note: This option requires that the snapshot directory is on the same hard disk
as the PureMessage files.

=item B<exclude>

Any files with this option set will be excluded from the snapshot process.  

=item B<delete_existing>

If this option is set (default is false), then local changes will be 
overwritten when restoring the file(s).

=back

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.

=head1 SEE ALSO

PureMessage::Snapshot




