package PureMessage::Snapshot::Utils;
use strict;
use Exporter;
use Digest::MD5 qw(md5);
use File::Copy;
use File::Path;
use PureMessage::Snapshot::Log;
our @ISA = qw(Exporter);
our @EXPORT_OK = qw(add_symlink add_link add_file add_dir Rename make_md5 
                    perms_changed restore_perms);

sub add_symlink {
    my ($orig, $dst, $new) = @_;
    symlink($dst, $new) || die("Can't symlink $dst to $new: $!");
    my ($uid, $gid) = (lstat($orig))[4,5];
    die("Can't lstat $orig: $!") unless defined $uid and defined $gid;
    PureMessage::_sys::lchown($uid, $gid, $new) || 
        die("Can't lchown $new (" . getpwuid($uid).":".getgrgid($gid) . "): $!");
    # symlinks don't need mode set
    ss_log("Created symlink ($dst, $new)", 2);
}

sub add_link {
    my ($orig, $dst, $new) = @_;
    if (link($dst, $new)) {
        ss_log("Hard linked $new to $dst", 2);
    }
    else {
        # Fallback to a copy
        copy($dst, $new) || die("Can't copy $dst to $new: $!");
        ss_log("Copied $dst to $new (hard link failed)", 2);
    }
    restore_perms($orig, $new);
}

sub add_file {
    my ($from, $to) = @_;
    copy($from, $to) || die("Can't copy $from to $to: $!");
    ss_log("Copied $from to $to", 2);
    restore_perms($from, $to);
}

sub add_dir {
    my ($from_base, $to_base, $dir) = @_;

    my @subdirs = split /\//, $dir;
    my $path = "";
    foreach my $d (@subdirs) {
        $path .= "$d/";
        my $to_path = "$to_base/$path";
        my $from_path = "$from_base/$path";
        next if -d $to_path or !-d $from_path;

        # Make the directory with the proper perms
        mkpath($to_path) || die("Can't mkpath $to_path: $!");
        restore_perms($from_path, $to_path);
        ss_log("Added directory ($to_path)", 2);
    }
}

# This function will unlink $new and then rename $old into place.  If the 
# rename fails, $new will not be restored.  The calling code should be 
# aware that the existing $new may be lost.
sub Rename {
    my ($old, $new) = @_;

    if (-e $new) {
        ss_log("Removed $new", 3);
        if ( ! -l $new && -d $new ) {
            rmtree($new) or die("Can't remove dir $new");
        }
        else {
            unlink($new) or die("Can't unlink file $new");
        }
    }

    if (rename($old, $new)) {
        ss_log("Renamed $old to $new", 3);
        return;
    }
    else {
        warn "Rename $old to $new failed: $!";
    }
    # rename failed - try a copy
    copy($old, $new) || die "Can't copy $old to $new: $!";
    ss_log("Copied $old to $new", 3);
}

sub make_md5 {
    my ($file) = @_;
    open(my $f, "<", $file) || die "Can't open $file: $!";
    binmode($f);
    return Digest::MD5->new->addfile($f)->hexdigest;
}

sub perms_changed {
    my ($fn1, $fn2) = @_;
    my @perms1 = (stat($fn1))[2,4,5];
    my @perms2 = (stat($fn2))[2,4,5];
    if ($perms1[0] != $perms2[0]) {
        ss_log(sprintf("Permissions differ ($fn1:%04o), ($fn2:%04o)",
                       $perms1[0] & 07777, $perms2[0] & 07777), 3);
        return 1;
    }
    elsif ($perms1[1] != $perms2[1] or $perms1[2] != $perms2[2]) {
        ss_log(sprintf("Ownerships differ ($fn1 \%s:\%s), ($fn2 \%s:\%s)", 
                       getpwuid($perms1[1]), getgrgid($perms1[2]),
                       getpwuid($perms2[1]), getgrgid($perms2[2]),), 3);
        return 1;
    }
    return 0;
}

sub restore_perms {
    my ($old, $new) = @_;
    # restore the permissions
    my ($perms, $uid, $gid) = (stat($old))[2,4,5];
    $perms &= 07777;
    chown($uid, $gid, $new) || 
        die("Can't chown $new (" . getpwuid($uid).":".getgrgid($gid) . "): $!");
    chmod($perms, $new) || die "Can't chmod $new: $!";
    ss_log(sprintf("Copied perms from $old to $new (\%s:\%s \%04o)",
                   (getpwuid($uid))[0], (getgrgid($gid))[0], $perms), 3);
}

1;


__END__

=head1 NAME

PureMessage::Snapshot::Utils

=head1 SYNOPSIS

File utility functions for dealing with snapshots

=head1 DESCRIPTION

The following functions are available:

=head1 AUTHOR

Sophos PLC

=head1 COPYRIGHT

Copyright (C) 2004, Sophos PLC
All Rights Reserved.
